"""
Sentinel-2 reader inherited from https://github.com/IPL-UV/DL-L8S2-UV.

It has several enhancements:
* Support for S2L2A images
* It can read directly images from a GCP bucket (for example data from  [here](https://cloud.google.com/storage/docs/public-datasets/sentinel-2))
* Windowed read and read and reproject in the same function (see `load_bands_bbox`)
* Creation of the image only involves reading one metadata file (`xxx.SAFE/MTD_{self.producttype}.xml`)


https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/document-library

"""
from rasterio import windows, features, coords
from rasterio.warp import reproject
from shapely.geometry import Polygon, MultiPolygon
from xml.etree.ElementTree import parse
import xml.etree.ElementTree as ET
import rasterio
import datetime
import io
from collections import OrderedDict
import numbers
import numpy as np
import os
import re
import tempfile
import sys
from typing import List, Tuple, Union


BANDS_S2 = ["B01", "B02","B03", "B04", "B05", "B06",
            "B07", "B08", "B8A", "B09", "B10", "B11", "B12"]

BANDS_RESOLUTION = OrderedDict({"B01": 60, "B02": 10,
                                "B03": 10, "B04": 10,
                                "B05": 20, "B06": 20,
                                "B07": 20, "B08": 10,
                                "B8A": 20, "B09": 60,
                                "B10": 60, "B11": 20, "B12": 20})


def process_metadata_msi(xml_file:str) -> Tuple[List[str], Polygon]:
    """
    Read the xml metadata and
    1) It gets the paths of the jp2 files of the SAFE product. It excludes the _TCI files.
    2) It gets the footprint of the image (polygon where image has valid values)
    Params:
        xml_file: description file found in root of .SAFE folder.  xxxx.SAFE/MTD_MSIL1C.xml or xxxx.SAFE/MTD_MSIL2A.xml

    Returns:
        List of jp2 files in the xml file
        Polygon with the extent of the image
    """
    if xml_file.startswith("gs://"):
        import fsspec
        fs = fsspec.filesystem("gs",requester_pays=True)
        with fs.open(xml_file, "rb") as file_obj:
            root = ET.fromstring(file_obj.read())
    else:
        root = ET.parse(xml_file).getroot()

    bands_elms = root.findall(".//IMAGE_FILE")
    jp2bands = [b.text+".jp2" for b in bands_elms if not b.text.endswith("_TCI")]

    footprint_txt = root.findall(".//EXT_POS_LIST")[0].text
    coords_split = footprint_txt.split(" ")[:-1]
    footprint = Polygon([(float(lngstr), float(latstr)) for latstr, lngstr in zip(coords_split[::2], coords_split[1::2])])

    return jp2bands, footprint


class S2Image:
    def __init__(self, s2_folder, out_res=10):
        mission, self.producttype, sensing_date_str, pdgs, relorbitnum, tile_number_field, product_discriminator = s2_name_split(
            s2_folder)

        out_res = int(out_res)
        assert out_res in {10, 20, 60}, "Not valid output resolution.Choose 10, 20, 60"

        # Default resolution to read
        self.out_res = out_res

        # Remove last trailing slash
        s2_folder = s2_folder[:-1] if s2_folder.endswith("/") else s2_folder
        self.name = os.path.basename(os.path.splitext(s2_folder)[0])

        self.datetime = datetime.datetime.strptime(sensing_date_str, "%Y%m%dT%H%M%S").replace(
            tzinfo=datetime.timezone.utc)
        self.folder = s2_folder
        self.metadata_msi = os.path.join(self.folder, f"MTD_{self.producttype}.xml").replace("\\","/")
        # load _pol from geometric_info product footprint!
        jp2bands, self._pol = process_metadata_msi(self.metadata_msi)

        self.granule_folder = os.path.dirname(os.path.dirname(jp2bands[0]))
        self.bands = list(BANDS_S2)

        self.band_check = None
        for band, res in BANDS_RESOLUTION.items():
            if res == self.out_res:
                self.band_check = band
                break

        self.granule = [os.path.join(self.folder, b).replace("\\","/") for b in jp2bands]

    def polygon(self):
        """ Footprint polygon of the image in lat/lng (epsg:4326) """
        return self._pol

    def _load_attrs(self):
        band_check_idx = self.bands.index(self.band_check)
        with rasterio.open(self.granule[band_check_idx], driver='JP2OpenJPEG') as src:
            self._shape = src.shape
            self._transform = src.transform
            self._crs = src.crs
            self._bounds = src.bounds
            self._dtype = src.dtypes[0]

    @property
    def dtype(self):
        if not hasattr(self, "_dtype"):
            self._load_attrs()
        return self._dtype

    @property
    def shape(self):
        if not hasattr(self, "_shape"):
            self._load_attrs()
        return self._shape

    @property
    def transform(self):
        if not hasattr(self, "_transform"):
            self._load_attrs()
        return self._transform

    @property
    def crs(self):
        if not hasattr(self, "_crs"):
            self._load_attrs()
        return self._crs

    @property
    def bounds(self):
        if not hasattr(self, "_bounds"):
            self._load_attrs()
        return self._bounds

    def __str__(self):
        return self.folder

    def load_bands_bbox(self, bbox_read, dst_crs=None, bands=None, resolution_dst_crs=None,
                        resampling=rasterio.warp.Resampling.cubic_spline) -> Tuple[np.ndarray, rasterio.Affine]:
        """
        Loads a list of bands in the bounding box provided at resolution_dst_crs

        :param bbox_read: bounding box to read in dst_crs coordinate reference system (CRS)
        :param dst_crs: dst crs e.g. {"init": "epsg:3857"}
        :param bands: id band to read from self.granule
        :param resolution_dst_crs: tuple or int with out resolution
        :param resampling: resampling method.

        :return: 3D numpy array of the reprojected data and out transform
        """
        if bands is None:
            bands = list(range(len(self.granule)))

        if dst_crs is None:
            dst_crs = self.crs
        else:
            assert resolution_dst_crs is not None, "Resolution must be set if dst_crs is set"

        if resolution_dst_crs is None:
            resolution_dst_crs = (self.out_res, self.out_res)

        dst_tuple =  read_tiffs_bbox_crs([self.granule[b] for b in bands],
                                         bbox_read, dst_crs,
                                         resolution_dst_crs=resolution_dst_crs, resampling=resampling,
                                         dtpye_dst=self.dtype)
        return dst_tuple

    def load_bands(self, bands=None, window=None):
        """
        window is a rasterio.windows.Window in self.out_res resolution

        :param bands: indexes of bands to read (self.granule[iband] for iband in bands)
        :param window: rasterio.windows.Window object
        :return: CHW array
        """
        if bands is None:
            bands = list(range(len(self.granule)))

        if window is None:
            shape = self.shape
            window = rasterio.windows.Window(col_off=0, row_off=0,
                                             width=shape[1], height=shape[0])

        # Use src.read if all bands have the same resolution == out_res
        if all(BANDS_RESOLUTION[self.bands[iband]] == self.out_res  for iband in bands):
            shape_window = windows.shape(window)
            dest_array = np.zeros((len(bands), ) + shape_window, dtype=self.dtype)
            for _i, iband in enumerate(bands):
                with rasterio.open(self.granule[iband]) as src:
                    dest_array[_i] = src.read(1, window=window, boundless=True)

            if window is not None:
                transform = rasterio.windows.transform(window, self.transform)
            else:
                transform = self.transform
            return dest_array, transform

        # Get bounding box to read
        bbox = windows.bounds(window, self.transform)
        return self.load_bands_bbox(bbox, bands=bands)

    def load_mask(self, window=None):
        band_check_idx = self.bands.index(self.band_check)
        band = self.load_bands(bands=[band_check_idx], window=window)
        return (band == 0) | (band == (2**16)-1)


class S2ImageL2A(S2Image):
    def __init__(self, s2_folder, out_res=10):
        super(S2ImageL2A, self).__init__(s2_folder, out_res=out_res)
        assert self.producttype == "MSIL2A", f"Unexpected product type {self.producttype} in image {self.folder}"

        # see https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2_SR#bands for a description of the granules data

        self.all_granules = list(self.granule)

        # Filter bands in self.all_granules according to the out_res
        self.granule = []
        self.bands = []
        for b in BANDS_S2:
            if b == "B10":
                continue
            res_band = BANDS_RESOLUTION[b]
            res_band = res_band if out_res < res_band else out_res
            band_and_res = f"{b}_{res_band}m.jp2"
            granules_match = [g for g in self.all_granules if g.endswith(band_and_res)]
            assert len(granules_match) == 1, f"Granules that match {band_and_res}: {granules_match}"
            self.granule.append(granules_match[0])
            self.bands.append(b)

        # Use SCL band for clouds
        res_band = 20 if out_res < 20 else out_res
        band_and_res = f"SCL_{res_band}m.jp2"
        granules_match = [g for g in self.all_granules if g.endswith(band_and_res)]
        assert len(granules_match) == 1, f"Granules that match {band_and_res}: {granules_match}"
        self.slc_granule = granules_match[0]

    def load_scl_bbox(self, bbox_read, dst_crs=None,resolution_dst_crs=None):
        if dst_crs is None:
            dst_crs = self.crs
        else:
            assert resolution_dst_crs is not None, "Resolution must be set if dst_crs is set"

        if resolution_dst_crs is None:
            resolution_dst_crs = (self.out_res, self.out_res)

        slc_tuple = read_tiffs_bbox_crs([self.slc_granule],
                                        bbox_read, dst_crs, resampling=rasterio.warp.Resampling.nearest,
                                        resolution_dst_crs=resolution_dst_crs,dtpye_dst=np.uint8)

        return slc_tuple


class S2ImageL1C(S2Image):
    def __init__(self, s2_folder, out_res=10, read_metadata=False):
        super(S2ImageL1C,self).__init__(s2_folder, out_res=out_res)

        assert self.producttype == "MSIL1C", f"Unexpected product type {self.producttype} in image {self.folder}"

        self.all_granules = list(self.granule)

        self.msk_clouds_file = os.path.join(self.folder, self.granule_folder, "MSK_CLOUDS_B00.gml").replace("\\","/")
        self.metadata_tl = os.path.join(self.folder, self.granule_folder, "MTD_TL.xml").replace("\\","/")

        self.bands = list(BANDS_S2)

        if read_metadata:
            self.read_metadata()

        self.granule = self.all_granules

        assert len(self.bands) == len(self.granule), f"Expected same but found different {len(self.bands)} {len(self.granule)}"

        # Granule in L1C does not include TCI
        # Assert bands in self.granule are ordered as in BANDS_S2
        assert len(self.granule) == len(BANDS_S2), f"Unexpected number of granules {len(self.granule)} expected {len(BANDS_S2)}"
        assert all(granule[-7:-4] == bname for bname, granule in zip(BANDS_S2, self.granule)), f"some granules are not in the expected order {self.granule}"

    def read_metadata(self):
        '''
        Read metadata TILE to parse information about the acquisition and properties of GRANULE bands
        Source: fmask.sen2meta.py
        :return: relevant attributes
        '''
        with open(self.metadata_tl) as f:
            root = ET.fromstring(f.read())
            # Stoopid XML namespace prefix
            nsPrefix = root.tag[:root.tag.index('}') + 1]
            nsDict = {'n1': nsPrefix[1:-1]}

            generalInfoNode = root.find('n1:General_Info', nsDict)
            # N.B. I am still not entirely convinced that this SENSING_TIME is really
            # the acquisition time, but the documentation is rubbish.
            sensingTimeNode = generalInfoNode.find('SENSING_TIME')
            sensingTimeStr = sensingTimeNode.text.strip()
            self.datetime = datetime.datetime.strptime(sensingTimeStr, "%Y-%m-%dT%H:%M:%S.%fZ")
            tileIdNode = generalInfoNode.find('TILE_ID')
            tileIdFullStr = tileIdNode.text.strip()
            self.tileId = tileIdFullStr.split('_')[-2]
            self.satId = tileIdFullStr[:3]
            self.procLevel = tileIdFullStr[13:16]  # Not sure whether to use absolute pos or split by '_'....

            geomInfoNode = root.find('n1:Geometric_Info', nsDict)
            geocodingNode = geomInfoNode.find('Tile_Geocoding')
            epsgNode = geocodingNode.find('HORIZONTAL_CS_CODE')
            self.epsg = epsgNode.text.split(':')[1]

            # Dimensions of images at different resolutions.
            self.dimsByRes = {}
            sizeNodeList = geocodingNode.findall('Size')
            for sizeNode in sizeNodeList:
                res = sizeNode.attrib['resolution']
                nrows = int(sizeNode.find('NROWS').text)
                ncols = int(sizeNode.find('NCOLS').text)
                self.dimsByRes[res] = (nrows, ncols)

            # Upper-left corners of images at different resolutions. As far as I can
            # work out, these coords appear to be the upper left corner of the upper left
            # pixel, i.e. equivalent to GDAL's convention. This also means that they
            # are the same for the different resolutions, which is nice.
            self.ulxyByRes = {}
            posNodeList = geocodingNode.findall('Geoposition')
            for posNode in posNodeList:
                res = posNode.attrib['resolution']
                ulx = float(posNode.find('ULX').text)
                uly = float(posNode.find('ULY').text)
                self.ulxyByRes[res] = (ulx, uly)

            # Sun and satellite angles.
            tileAnglesNode = geomInfoNode.find('Tile_Angles')
            sunZenithNode = tileAnglesNode.find('Sun_Angles_Grid').find('Zenith')
            self.angleGridXres = float(sunZenithNode.find('COL_STEP').text)
            self.angleGridYres = float(sunZenithNode.find('ROW_STEP').text)
            self.sunZenithGrid = self.makeValueArray(sunZenithNode.find('Values_List'))
            sunAzimuthNode = tileAnglesNode.find('Sun_Angles_Grid').find('Azimuth')
            self.sunAzimuthGrid = self.makeValueArray(sunAzimuthNode.find('Values_List'))
            self.anglesGridShape = self.sunAzimuthGrid.shape

            # Now build up the viewing angle per grid cell, from the separate layers
            # given for each detector for each band. Initially I am going to keep
            # the bands separate, just to see how that looks.
            # The names of things in the XML suggest that these are view angles,
            # but the numbers suggest that they are angles as seen from the pixel's
            # frame of reference on the ground, i.e. they are in fact what we ultimately want.
            viewingAngleNodeList = tileAnglesNode.findall('Viewing_Incidence_Angles_Grids')
            self.viewZenithDict = self.buildViewAngleArr(viewingAngleNodeList, 'Zenith')
            self.viewAzimuthDict = self.buildViewAngleArr(viewingAngleNodeList, 'Azimuth')

            # Make a guess at the coordinates of the angle grids. These are not given
            # explicitly in the XML, and don't line up exactly with the other grids, so I am
            # making a rough estimate. Because the angles don't change rapidly across these
            # distances, it is not important if I am a bit wrong (although it would be nice
            # to be exactly correct!).
            (ulx, uly) = self.ulxyByRes["10"]
            self.anglesULXY = (ulx - self.angleGridXres / 2.0, uly + self.angleGridYres / 2.0)

    def buildViewAngleArr(self, viewingAngleNodeList, angleName):
        """
        Build up the named viewing angle array from the various detector strips given as
        separate arrays. I don't really understand this, and may need to re-write it once
        I have worked it out......

        The angleName is one of 'Zenith' or 'Azimuth'.
        Returns a dictionary of 2-d arrays, keyed by the bandId string.
        """
        angleArrDict = {}
        for viewingAngleNode in viewingAngleNodeList:
            bandId = viewingAngleNode.attrib['bandId']
            angleNode = viewingAngleNode.find(angleName)
            angleArr = self.makeValueArray(angleNode.find('Values_List'))
            if bandId not in angleArrDict:
                angleArrDict[bandId] = angleArr
            else:
                mask = (~np.isnan(angleArr))
                angleArrDict[bandId][mask] = angleArr[mask]
        return angleArrDict

    def get_polygons_bqa(self):
        def polygon_from_coords(coords, fix_geom=False, swap=True, dims=2):
            """
            Return Shapely Polygon from coordinates.
            - coords: list of alterating latitude / longitude coordinates
            - fix_geom: automatically fix geometry
            """
            assert len(coords) % dims == 0
            number_of_points = int(len(coords) / dims)
            coords_as_array = np.array(coords)
            reshaped = coords_as_array.reshape(number_of_points, dims)
            points = [(float(i[1]), float(i[0])) if swap else ((float(i[0]), float(i[1]))) for i in reshaped.tolist()]
            polygon = Polygon(points).buffer(0)
            try:
                assert polygon.is_valid
                return polygon
            except AssertionError:
                if fix_geom:
                    return polygon.buffer(0)
                else:
                    raise RuntimeError("Geometry is not valid.")


        exterior_str = str("eop:extentOf/gml:Polygon/gml:exterior/gml:LinearRing/gml:posList")
        interior_str = str("eop:extentOf/gml:Polygon/gml:interior/gml:LinearRing/gml:posList")
        root = parse(self.msk_clouds_file).getroot()
        nsmap = {k: v for k, v in root.nsmap.items() if k}
        try:
            for mask_member in root.iterfind("eop:maskMembers", namespaces=nsmap):
                for feature in mask_member:
                    type = feature.findtext("eop:maskType", namespaces=nsmap)

                    ext_elem = feature.find(exterior_str, nsmap)
                    dims = int(ext_elem.attrib.get('srsDimension', '2'))
                    ext_pts = ext_elem.text.split()
                    exterior = polygon_from_coords(ext_pts, fix_geom=True, swap=False, dims=dims)
                    try:
                        interiors = [polygon_from_coords(int_pts.text.split(), fix_geom=True, swap=False, dims=dims)
                                     for int_pts in feature.findall(interior_str, nsmap)]
                    except AttributeError:
                        interiors = []

                    yield dict(geometry=Polygon(exterior, interiors).buffer(0),
                               attributes=dict(maskType=type),
                               interiors=interiors)

        except StopIteration:
            yield dict(geometry=Polygon(),
                       attributes=dict(maskType=None),
                       interiors=[])
            raise StopIteration()

    def load_clouds_bqa(self, window=None):
        mask_types = ["OPAQUE", "CIRRUS"]
        poly_list = list(self.get_polygons_bqa())

        nrows, ncols = self.shape
        transform_ = self.transform

        def get_mask(mask_type=mask_types[0]):
            assert mask_type in mask_types, "mask type must be OPAQUE or CIRRUS"
            fill_value = {m: i+1 for i, m in enumerate(mask_types)}
            n_polys = np.sum([poly["attributes"]["maskType"] == mask_type for poly in poly_list])
            msk = np.zeros(shape=(nrows, ncols), dtype=np.float32)
            if n_polys > 0:
                # n_interiors = np.sum([len(poly) for poly in poly_list if poly["interiors"]])
                multi_polygon = MultiPolygon([poly["geometry"]
                                              for poly in poly_list
                                              if poly["attributes"]["maskType"] == mask_type]).buffer(0)
                bounds = multi_polygon.bounds
                bbox2read = coords.BoundingBox(*bounds)
                window_read = windows.from_bounds(*bbox2read, transform_)
                slice_read = tuple(slice(int(round(s.start)), int(round(s.stop))) for s in window_read.toslices())
                out_shape = tuple([s.stop - s.start for s in slice_read])
                transform_slice = windows.transform(window_read, transform_)

                shapes = [({"type": "Polygon",
                            "coordinates": [np.stack([
                                p_elem["geometry"].exterior.xy[0],
                                p_elem["geometry"].exterior.xy[1]], axis=1).tolist()]}, fill_value[mask_type])
                          for p_elem in poly_list if p_elem["attributes"]['maskType'] == mask_type]
                sub_msk = features.rasterize(shapes=shapes, fill=0,
                                             out_shape=out_shape, dtype=np.float32,
                                             transform=transform_slice)
                msk[slice_read] = sub_msk

            return msk

        if window is None:
            shape = self.shape
            window = rasterio.windows.Window(col_off=0, row_off=0,
                                             width=shape[1], height=shape[0])

        mask = self.load_mask(window=window)

        slice_ = window.toslices()

        msk_op_cirr = [np.ma.MaskedArray(get_mask(mask_type=m)[slice_], mask=mask) for m in mask_types]
        msk_clouds = np.ma.MaskedArray(np.clip(np.sum(msk_op_cirr, axis=0), 0, 1), mask=mask)
        return msk_clouds

    @staticmethod
    def makeValueArray(valuesListNode):
        """
        Take a <Values_List> node from the XML, and return an array of the values contained
        within it. This will be a 2-d numpy array of float32 values (should I pass the dtype in??)

        """
        valuesList = valuesListNode.findall('VALUES')
        vals = []
        for valNode in valuesList:
            text = valNode.text
            vals.append([np.float32(x) for x in text.strip().split()])

        return np.array(vals)


def s2loader(s2folder:str, out_res:int=10) -> Union[S2ImageL2A, S2ImageL1C]:
    """
    Loads a S2ImageL2A or S2ImageL1C depending on the product type

    :param s2folder: .SAFE folder. Expected standard ESA naming convention (see s2_name_split fun)
    :param out_res: default output resolution

    """
    _, producttype_nos2, _, _, _, _, _ = s2_name_split(s2folder)
    if producttype_nos2 == "MSIL2A":
        return S2ImageL2A(s2folder, out_res=out_res)
    elif producttype_nos2 == "MSIL1C":
        return S2ImageL1C(s2folder, out_res=out_res)

    raise NotImplementedError(f"Don't know how to load {producttype_nos2} products")


NEW_FORMAT = "(S2\w{1})_(MSIL\w{2})_(\d{4}\d{2}\d{2}T\d{2}\d{2}\d{2})_(\w{5})_(\w{4})_T(\w{5})_(\w{15})"
OLD_FORMAT = "(S2\w{1})_(\w{4})_(\w{3}_\w{6})_(\w{4})_(\d{8}T\d{6})_(\w{4})_V(\d{4}\d{2}\d{2}T\d{6})_(\d{4}\d{2}\d{2}T\d{6})"

def s2_name_split(s2file):
    """
    https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/naming-convention

    ```
    s2l1c = "S2A_MSIL1C_20151218T182802_N0201_R127_T11SPD_20151218T182756.SAFE"
    mission, producttype, sensing_date_str, pdgs, relorbitnum, tile_number_field, product_discriminator = s2_name_split(s2l1c)
    ```

    S2A_MSIL1C_20151218T182802_N0201_R127_T11SPD_20151218T182756.SAFE
    MMM_MSIXXX_YYYYMMDDTHHMMSS_Nxxyy_ROOO_Txxxxx_<Product Discriminator>.SAFE
    MMM: is the mission ID(S2A/S2B)
    MSIXXX: MSIL1C denotes the Level-1C product level/ MSIL2A denotes the Level-2A product level
    YYYYMMDDHHMMSS: the datatake sensing start time
    Nxxyy: the PDGS Processing Baseline number (e.g. N0204)
    ROOO: Relative Orbit number (R001 - R143)
    Txxxxx: Tile Number field
    SAFE: Product Format (Standard Archive Format for Europe)

    :param s2l1c:
    :return:
    """
    s2file = s2file[:-1] if s2file.endswith("/") else s2file
    basename = os.path.basename(os.path.splitext(s2file)[0])
    matches = re.match(NEW_FORMAT, basename)
    if matches is not None:
        return matches.groups()


def s2_old_format_name_split(s2file):
    """
    https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/naming-convention

    ```
    s2l1c = "S2A_OPER_PRD_MSIL1C_PDMC_20151206T093912_R090_V20151206T043239_20151206T043239.SAFE"
    mission, opertortest, filetype, sitecenter,  creation_date_str, relorbitnum, sensing_time_start, sensing_time_stop = s2_old_format_name_split(s2l1c)
    ```

    :param s2l1c:
    :return:
    """
    s2file = s2file[:-1] if s2file.endswith("/") else s2file
    basename = os.path.basename(os.path.splitext(s2file)[0])
    matches = re.match(OLD_FORMAT, basename)
    if matches is not None:
        return matches.groups()


def pad_window(window, pad_size):
    return rasterio.windows.Window(window.col_off - pad_size[1],
                                   window.row_off - pad_size[0],
                                   width=window.width + 2*pad_size[1],
                                   height=window.height + 2*pad_size[1])

def round_window(window):
    return window.round_lengths(op="ceil").round_offsets(op="floor")


def read_tiffs_bbox_crs(tiff_files, bbox_read, dst_crs, resolution_dst_crs, resampling, dtpye_dst=np.float32):
    """
    Read data from a list of tiffs. Read the data in bbox_read reprojected to dst_crs with resolution resolution_dst_crs.

    :param tiff_files:
    :param bbox_read: Bounding box to read in dst_crs CRS
    :param dst_crs:
    :param resampling:
    :param resolution_dst_crs:
    :param dtpye_dst:
    :return:
    """

    if isinstance(resolution_dst_crs, numbers.Number):
        resolution_dst_crs = (abs(resolution_dst_crs), abs(resolution_dst_crs))

    # Compute affine transform out crs
    dst_transform = transform_bounds_res(bbox_read, resolution_dst_crs)
    # Compute size of window in out crs
    window_out = rasterio.windows.from_bounds(*bbox_read,
                                              transform=dst_transform).round_lengths(op="ceil")
    # Create out array
    destination = np.zeros((len(tiff_files), window_out.height, window_out.width),
                           dtype=dtpye_dst)

    for i, b in enumerate(tiff_files):
        try:
            np_array_in, transform_in, src_crs = read_overlap(b, bbox_read, dst_crs)
        except Exception as e:
            if b.startswith("gs://"):
                print(f"\tError reading remote file {sys.exc_info()[0]}. \n Copying file to local and trying again")
                from google.cloud import storage
                client = storage.Client()
                with tempfile.NamedTemporaryFile() as tmpfile:
                    client.download_blob_to_file(b, tmpfile)
                    np_array_in, transform_in, src_crs = read_overlap(tmpfile.name, bbox_read, dst_crs)
            else:
                raise e

        np_array_in = np_array_in.astype(dtpye_dst)

        rasterio.warp.reproject(
            np_array_in,
            destination[i],
            src_transform=transform_in,
            src_crs=src_crs,
            dst_transform=dst_transform,
            dst_crs=dst_crs,
            src_nodata=0.,
            dst_nodata=0.,
            resampling=resampling)

    return destination, dst_transform


def read_overlap(tiff_file, bbox_read, dst_crs, pad_add=(3, 3)):
    with rasterio.open(tiff_file) as src_b:
        bounds_in = rasterio.warp.transform_bounds(dst_crs, src_b.crs, *bbox_read)
        window_in = rasterio.windows.from_bounds(*bounds_in, transform=src_b.transform)
        window_in = pad_window(window_in, pad_add)  # Add padding for bicubic int
        window_in = round_window(window_in)
        transform_in = rasterio.windows.transform(window_in, transform=src_b.transform)

        src_crs = src_b.crs
        np_array_in = src_b.read(1, window=window_in, boundless=True)
    return np_array_in, transform_in, src_crs



def mosaic_s2(s2objs, bbox_read, dst_crs, res_s2, threshold_stop_invalid=1/1000., sort=True):
    """
    Mosaic a list of s2objs. Reorder s2objs by largest overlap

    :param s2objs: Could be L2A or L1C S2 products.
    :param bbox_read:
    :param dst_crs:
    :param res_s2:
    :param threshold_stop_invalid:
    :param sort: Sort s2ojs
    :return:
    """
    # Order objects by overlap with pol_read
    bbox_lnglat = rasterio.warp.transform_bounds(dst_crs,
                                                 {'init': 'epsg:4326'},
                                                 *bbox_read)
    pol_read = Polygon(generate_polygon(bbox_lnglat))

    if sort:
        s2objs = sorted(s2objs, key= lambda s2: s2.polygon().intersection(pol_read).area/pol_read.area,
                        reverse=True)

    invalid_destination = None
    current_clouds = None
    for s2obj in s2objs:
        print(f"{s2obj.name} reprojecting")
        destination_iter, dst_transform = s2obj.load_bands_bbox(bbox_read, dst_crs=dst_crs,
                                                                resolution_dst_crs=(res_s2, res_s2))

        if invalid_destination is None:
            destination_array = destination_iter
            if s2obj.producttype == "MSIL2A":
                current_clouds, _ = s2obj.load_scl_bbox(bbox_read, dst_crs=dst_crs,
                                                        resolution_dst_crs=(res_s2, res_s2))
                current_clouds = current_clouds[0]
        else:
            destination_array[:, invalid_destination] = destination_iter[:, invalid_destination]
            if s2obj.producttype == "MSIL2A":
                slc, _ = s2obj.load_scl_bbox(bbox_read, dst_crs=dst_crs, resolution_dst_crs=(res_s2, res_s2))
                slc = slc[0]
                current_clouds[invalid_destination] = slc[invalid_destination]

        if s2obj.producttype == "MSIL2A":
            invalid_destination = current_clouds <= 1
        else:
            invalid_destination = np.all(destination_array < 1e-6, axis=0, keepdims=False) |\
                                  np.all(destination_array == (2 ** 16) - 1, axis=0, keepdims=False)

        frac_invalids = np.sum(invalid_destination) / np.prod(invalid_destination.shape)

        if frac_invalids < threshold_stop_invalid:
            break

    return destination_array, current_clouds


def generate_polygon(bbox):
    """
    Generates a list of coordinates: [[x1,y1],[x2,y2],[x3,y3],[x4,y4],[x1,y1]]
    """
    return [[bbox[0],bbox[1]],
             [bbox[2],bbox[1]],
             [bbox[2],bbox[3]],
             [bbox[0],bbox[3]],
             [bbox[0],bbox[1]]]

def transform_bounds_res(bbox_read, resolution_dst_crs):
    """ Compute affine transform for a given bounding box and resolution. bbox_read and resolution_dst_crs are expected in the same CRS"""
    # Compute affine transform out crs
    return rasterio.transform.from_origin(min(bbox_read[0], bbox_read[2]),
                                          max(bbox_read[1], bbox_read[3]),
                                          resolution_dst_crs[0], resolution_dst_crs[1])



class AttrDict(dict):
    """ Dictionary subclass whose entries can be accessed like attributes
        (as well as normally).
    """

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    @staticmethod
    def from_nested_dicts(data):
        """ Construct nested AttrDicts from nested dictionaries. """
        if not isinstance(data, dict):
            return data
        else:
            return AttrDict({key: AttrDict.from_nested_dicts(data[key])
                             for key in data})


import torch.nn as nn

from ml4floods.models.architectures import layer_factory


class SimpleLinear(nn.Module):
    """
    Linear model
    """
    def __init__(self, n_channels, n_class):
        super().__init__()

        self.conv = nn.Sequential(
            nn.Conv2d(n_channels, n_class, 1)
        )

    def forward(self, x):
        
        res = self.conv(x)
        
        return res


class SimpleCNN(nn.Module):
    """
    5-layer fully conv CNN
    """
    def __init__(self, n_channels, n_class):
        super().__init__()

        self.conv = nn.Sequential(
            layer_factory.double_conv(n_channels, 64),
            layer_factory.double_conv(64, 128),
            nn.Conv2d(128, n_class, 1)
        )
        
    def forward(self, x):
        
        res = self.conv(x)
        
        return res



import torch.nn as nn


def double_conv(in_channels, out_channels):
    return nn.Sequential(
        nn.Conv2d(in_channels, out_channels, 3, padding=1),
        nn.ReLU(inplace=True),
        nn.Conv2d(out_channels, out_channels, 3, padding=1),
        nn.ReLU(inplace=True)
    )


from typing import Optional

import torch
import torch.nn.functional as F
from typing import Optional, List


def dice_loss_mask_invalid(logits: torch.Tensor, target:torch.Tensor, smooth=1.) -> float:
    """
    Dice loss masking invalids (it masks the 0 value in the target tensor)

    Args:
        logits: (B, C, H, W) tensor with logits (no softmax)
        target: (B, H, W) tensor. int values in {0,...,C} it considers 0 to be the invalid value
        smooth: Value to avoid div by zero

    Returns:
        averaged loss (float)

    """
    assert logits.dim() == 4, f"Expected 4D tensor logits"
    assert target.dim() == 3, f"Expected 3D tensor target"

    pred = torch.softmax(logits, dim=1)
    valid = (target != 0) # (B, H, W) tensor
    target_without_invalids = (target - 1) * valid  # Set invalids to land

    target_one_hot_without_invalid = torch.nn.functional.one_hot(target_without_invalids,
                                                                 num_classes=pred.shape[1]).permute(0, 3, 1, 2)
    axis_red = (2, 3) # H, W reduction

    pred_valid = pred * valid.unsqueeze(1).float()  # # Set invalids to 0 (all values in prob tensor are 0

    intersection = (pred_valid * target_one_hot_without_invalid).sum(dim=axis_red) # (B, C) tensor

    union = pred_valid.sum(dim=axis_red) + target_one_hot_without_invalid.sum(dim=axis_red)  # (B, C) tensor

    dice_score = ((2. * intersection + smooth) /
                 (union + smooth))

    loss = (1 - dice_score)  # (B, C) tensor

    return torch.mean(loss)

def cross_entropy_loss_mask_invalid(logits: torch.Tensor, target:torch.Tensor, weight:Optional[torch.Tensor]=None) -> float:
    """
    F.cross_entropy loss masking invalids (it masks the 0 value in the target tensor)

    Args:
        logits: (B, C, H, W) tensor with logits (no softmax!)
        target: (B, H, W) tensor. int values in {0,...,C} it considers 0 to be the invalid value
        weight: (C, ) tensor. weight per class value to cross_entropy function

    Returns:
        averaged loss

    """
    assert logits.dim() == 4, f"Expected 4D tensor logits"
    assert target.dim() == 3, f"Expected 3D tensor target"

    valid = (target != 0)
    target_without_invalids = (target - 1) * valid

    # BCE Loss (ignoring invalid values)
    bce = F.cross_entropy(logits, target_without_invalids,
                          weight=weight, reduction='none')  # (B, 1, H, W)

    bce *= valid  # mask out invalid pixels

    return torch.sum(bce / (torch.sum(valid) + 1e-6))


def calc_loss_mask_invalid(logits: torch.Tensor, target:torch.Tensor,
                           bce_weight:float=0.5, weight:Optional[torch.Tensor]=None) -> float:
    """
    Weighted BCE and Dice loss masking invalids:
     bce_loss * bce_weight + dice_loss * (1-bce_weight)
    Args:
        logits: (B, C, H, W) tensor with logits (no softmax!)
        target: (B, H, W) tensor. int values in {0,...,C} it considers 0 to be the invalid value
        bce_weight: weight of the bce part
        weight: (C, ) tensor. weight per class value to cross_entropy function

    Returns:

    """

    bce = cross_entropy_loss_mask_invalid(logits, target, weight=weight)

    # Dice Loss
    # Perform spatial softmax over NxCxHxW
    dice = dice_loss_mask_invalid(logits, target) # (B, C)

    # Weighted sum
    return bce * bce_weight + dice * (1 - bce_weight)


def binary_cross_entropy_loss_mask_invalid(logits: torch.Tensor, target: torch.Tensor,
                                           pos_weight:Optional=None) -> float:
    """

    Args:
        logits: (B, H, W) tensor with logits (no signmoid!)
        target: (B, H, W) Tensor with values encoded as {0: invalid, 1: neg_xxx, 2:pos_xxx}
        pos_weight: weight of positive class

    Returns:

    """
    assert logits.dim() == 3, f"Unexpected shape of logits. Logits: {logits.shape} target: {target.shape}"
    assert target.dim() == 3, f"Unexpected shape of target. Logits: {logits.shape} target: {target.shape}"

    valid = (target != 0)
    target_without_invalids = (target - 1) * valid

    pixelwise_bce = F.binary_cross_entropy_with_logits(logits, target_without_invalids.float(), reduction='none',
                                                       pos_weight=pos_weight)

    pixelwise_bce *= valid  # mask out invalid pixels

    return torch.sum(pixelwise_bce / (torch.sum(valid) + 1e-6))

def calc_loss_multioutput_logistic_mask_invalid(logits: torch.Tensor, target: torch.Tensor,
                                                pos_weight_problem:Optional[List[float]]=None,
                                                weight_problem:Optional[List[float]]=None) -> float:

    assert logits.dim() == 4, "Unexpected shape of logits"
    assert target.dim() == 4, "Unexpected shape of target"

    if weight_problem is None:
        weight_problem = [ 1/logits.shape[1] for _ in range(logits.shape[1])]

    total_loss = 0
    for i in range(logits.shape[1]):
        pos_weight = torch.tensor(pos_weight_problem[i], device=logits.device) if pos_weight_problem is not None else None
        curr_loss = binary_cross_entropy_loss_mask_invalid(logits[:, i], target[:, i], pos_weight=pos_weight)
        total_loss += curr_loss*weight_problem[i]

    return total_loss






import traceback
import warnings

import ee
import time
import os
from glob import glob
from typing import Optional, Callable, List, Tuple, Union
from shapely.geometry import mapping, Polygon
import numpy as np
import geopandas as gpd
import pandas as pd
import fsspec
from datetime import datetime, timezone
import math

BANDS_NAMES = {
    # Sentinel-2 L1C COPERNICUS/S2 COPERNICUS/S2_HARMONIZED
    "COPERNICUS/S2_HARMONIZED" : ["B1","B2","B3","B4", "B5", "B6", "B7", "B8", "B8A", "B9", "B10", "B11", "B12", "QA60"],
    # Sentinel-2 L2A COPERNICUS/S2_SR
    "COPERNICUS/S2_SR_HARMONIZED" : ["B1","B2","B3","B4", "B5", "B6", "B7", "B8", "B8A", "B9", "B11", "B12", "SCL"],
    # Landsat-8/9 TOA
    "Landsat" : ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "B9", "B10", "B11", "QA_PIXEL"]
}


def _get_collection(collection_name, date_start, date_end, bounds):
    collection = ee.ImageCollection(collection_name)
    collection_filtered = collection.filterDate(date_start, date_end) \
        .filterBounds(bounds)

    n_images = int(collection_filtered.size().getInfo())

    return collection_filtered, n_images

    # add cloud probability
def add_cloud_prob_landsat(img:ee.Image) -> ee.Image:
    bqa = img.select(["QA_PIXEL"], ["probability"])
    clouds = bqa.bitwise_and(int("0000000000001000",2)).gt(0).multiply(100).toUint16()

    # Store images in S2 range
    img_radiances = img.select(BANDS_NAMES["Landsat"][:-1]).multiply(10_000).toUint16()
    img_return = img_radiances.addBands(img.select("QA_PIXEL")).addBands(clouds)
    img_return = img_return.copyProperties(img)
    img_return = img_return.set("system:time_start", img.get("system:time_start"))
    return img_return

def get_landsat_collection(date_start:datetime, date_end:datetime,
                           bounds:ee.Geometry,
                           verbose:int=1) -> Optional[ee.ImageCollection]:
    """
    Returns a Landsat-8 and Landsat-9 time series of daily mosaicked images over the bounds. It adds a probability band
    to mimic the S2 case; additionally values are 10_000 and casted to uint16.
    (also to maintain the similarity with S2 collections).

    Args:
        date_start: start search date
        date_end: end search date
        bounds: polygon with the AoI to download
        verbose: print stuff

    Returns:
        ImageCollection

    """
    # GEE doesnt like time zones
    date_start = date_start.replace(tzinfo=None)
    date_end = date_end.replace(tzinfo=None)
    l8 = ee.ImageCollection("LANDSAT/LC08/C02/T1_RT_TOA").filterDate(date_start, date_end).filterBounds(bounds)
    l9 = ee.ImageCollection("LANDSAT/LC09/C02/T1_TOA").filterDate(date_start, date_end).filterBounds(bounds)
    l89 = l8.merge(l9)
    n_images = int(l89.size().getInfo())
    if n_images <= 0:
        if verbose > 1:
            print(f"Not images found for collection LANDSAT/LC08/C02/T1_RT_TOA and LANDSAT/LC09/C02/T1_TOA date start: {date_start} date end: {date_end}")
        return

    l89 = l89.map(add_cloud_prob_landsat)

    daily_mosaic = collection_mosaic_day(l89, bounds)
    # fun_before_mosaic=lambda img: img.toFloat().resample("bicubic")) # Bicubic resampling for 60m res bands?

    # Add cloud info and valid info
    count_fun = get_count_function(["QA_PIXEL"], bounds)
    daily_mosaic = daily_mosaic.map(count_fun)

    return daily_mosaic

def add_probability_from_BQA(img:ee.Image) -> ee.Image:
    """
    Adds to the image the cloud probability in [0,100] computed from BQA layer

    # Code taken from https://github.com/IPL-UV/ee_ipl_uv

    var qa = image.select('QA60');
    // Bits 10 and 11 are clouds and cirrus, respectively.
    var cloudBitMask = 1 << 10;
    var cirrusBitMask = 1 << 11;

    // Both flags should be set to zero, indicating clear conditions.
    var mask = qa.bitwiseAnd(cloudBitMask).eq(0)
                 .and(qa.bitwiseAnd(cirrusBitMask).eq(0));
    """
    qa = img.select(['QA60'], ["probability"])
    cloud_bit_mask =  1 << 10 # int("0000010000000000", 2)
    cirrus_bit_mask = int("0000100000000000", 2)
    # qa.bitwiseAnd(cloud_bit_mask).gt(0).Or(qa.bitwiseAnd(cirrus_bit_mask).gt(0))

    # Ignore cirrus since we can predict on water in those areas
    cloud_mask = qa.bitwiseAnd(cloud_bit_mask).gt(0).multiply(100).toUint16()
    return img.addBands(cloud_mask)


def get_s2_collection(date_start:datetime, date_end:datetime,
                      bounds:ee.Geometry,
                      collection_name:str="COPERNICUS/S2_HARMONIZED", bands:Optional[List[str]]=None,
                      force_s2cloudless:bool=True,
                      verbose:int=1) -> Optional[ee.ImageCollection]:
    """
    Returns an ee.ImageCollection with mosaicked S2 images joined with the s2cloudless cloud masks
    (see collection COPERNICUS/S2_CLOUD_PROBABILITY). This function also filters repeated images of the same date
    over the same location.

    Args:
        date_start: start search date
        date_end: end search date
        bounds: polygon with the AoI to download
        collection_name: "COPERNICUS/S2" for L1C Sentinel-2 images and ""COPERNICUS/S2_SR" for L2A images.
        bands: list of bands to get
        force_s2cloudless: if s2cloudless is missing for some S2 images it will skip those instead of raising
        not implemented errors.
        verbose: print stuff

    Returns:
        image collection with S2 imaages and s2cloudless cloud masks.
        None if no image satisfies the condition in the collection
    """

    # GEE doesnt like time zones
    date_start = date_start.replace(tzinfo=None)
    date_end = date_end.replace(tzinfo=None)
    img_col_all, n_images_col = _get_collection(collection_name, date_start, date_end, bounds)
    if n_images_col <= 0:
        if verbose > 1:
            print(f"Not images found for collection {collection_name} date start: {date_start} date end: {date_end}")
        return

    if bands is None:
        bands = BANDS_NAMES[collection_name]

    img_col_all = img_col_all.select(bands)

    # Import and filter s2cloudless.
    s2_cloudless_col = (ee.ImageCollection('COPERNICUS/S2_CLOUD_PROBABILITY')
        # .filterBounds(bounds)
        .filterDate(date_start, date_end))

    img_col_all_join = ee.ImageCollection(ee.Join.saveFirst('s2cloudless').apply(**{
        'primary': img_col_all,
        'secondary': s2_cloudless_col,
        'condition': ee.Filter.equals(**{
            'leftField': 'system:index',
            'rightField': 'system:index'
        })
    }))
    # Add s2cloudless as new band
    img_col_all_join = img_col_all_join.map(lambda x: x.addBands(ee.Image(x.get('s2cloudless')).select('probability')))

    # Check no image is lost in the inner join
    n_images_join = img_col_all_join.size().getInfo()

    if n_images_join < n_images_col:
        if not force_s2cloudless:
            # Compute cloud mask from BQA band and add it as probability layer
            warnings.warn(
                f"Not all the images in the S2 collection {n_images_col} have s2cloudless cloud mask {n_images_join}. WE WILL IGNORE CLOUDS!")
            img_col_all_join = img_col_all.map(add_probability_from_BQA)

        else:
            warnings.warn(
                f"Not all the images in the S2 collection {n_images_col} have s2cloudless cloud mask {n_images_join}. We will skip images with missing cloud masks")

    daily_mosaic =  collection_mosaic_day(img_col_all_join, bounds)
                                    #fun_before_mosaic=lambda img: img.toFloat().resample("bicubic")) # Bicubic resampling for 60m res bands?

    # Add cloud info and valid info
    count_fun = get_count_function(bands, bounds)
    daily_mosaic = daily_mosaic.map(count_fun)

    return daily_mosaic
    # Filter images with many invalids

def get_count_function(bands, bounds):
    def _count_valid_clouds(img:ee.Image) -> ee.Image:
        mascara_valids = img.mask()
        mascara_valids = mascara_valids.select(bands)
        mascara_valids = mascara_valids.reduce(ee.Reducer.allNonZero())
        dictio = mascara_valids.reduceRegion(reducer=ee.Reducer.mean(), geometry=bounds,
                                             bestEffort=True, scale=10.)

        img = img.set("valids", dictio.get("all"))

        # Count clouds
        cloud_probability = img.select("probability")
        dictio = cloud_probability.reduceRegion(reducer=ee.Reducer.mean(), geometry=bounds,
                                                bestEffort=True, scale=10.)

        img = img.set("cloud_probability", dictio.get("probability"))

        return img

    return _count_valid_clouds


def collection_mosaic_day(imcol:ee.ImageCollection, region_of_interest:ee.Geometry)-> ee.ImageCollection:
    """
    Groups by solar day the images in the image collection. This function is useful to discard repeated images in
    image collections for example in the case of Sentinel-2 images.

    Args:
        imcol: image collection
        region_of_interest: needed to find the solar day

    Returns:

    """
    # https://gis.stackexchange.com/questions/280156/mosaicking-a-image-collection-by-date-day-in-google-earth-engine
    imlist = imcol.toList(imcol.size())

    # longitude, latitude = region_of_interest.centroid().coordinates().getInfo()
    longitude = region_of_interest.centroid().coordinates().get(0)

    hours_add = ee.Number(longitude).multiply(12/180.)
    # solar_time = utc_time - hours_add

    unique_solar_dates = imlist.map(lambda im: ee.Image(im).date().advance(hours_add, "hour").format("YYYY-MM-dd")).distinct()

    def mosaic_date(solar_date_str):
        solar_date = ee.Date(solar_date_str)
        utc_date = solar_date.advance(hours_add.multiply(-1), "hour")

        ims_day = imcol.filterDate(utc_date, utc_date.advance(1, "day"))

        dates = ims_day.toList(ims_day.size()).map(lambda x: ee.Image(x).date().millis())
        median_date = dates.reduce(ee.Reducer.median())

        im = ims_day.mosaic()

        im = im.copyProperties(ee.Image(ims_day.first()))

        return im.set({
            "system:time_start": median_date,
            "system:id": solar_date.format("YYYY-MM-dd"),
            "system:index": solar_date.format("YYYY-MM-dd")
        })

    mosaic_imlist = unique_solar_dates.map(mosaic_date)
    return ee.ImageCollection(mosaic_imlist)


PROPERTIES_DEFAULT = ["system:index", "system:time_start"]
def img_collection_to_feature_collection(img_col:ee.ImageCollection,
                                         properties:List[str]=PROPERTIES_DEFAULT,
                                         as_geopandas:bool=False) -> Union[ee.FeatureCollection, gpd.GeoDataFrame]:
    """Transforms the image collection to a feature collection """

    properties = ee.List(properties)

    def extractFeatures(img):
        values = properties.map(lambda prop: img.get(prop))
        dictio = ee.Dictionary.fromLists(properties, values)
        return ee.Feature(img.geometry(), dictio)

    feature_collection = ee.FeatureCollection(img_col.map(extractFeatures))
    if as_geopandas:
        geodf = gpd.GeoDataFrame.from_features(feature_collection.getInfo(), crs="EPSG:4326")
        if "system:time_start" in geodf.columns:
            geodf["datetime"] = pd.to_datetime(geodf["system:time_start"],unit="ms")
        return geodf

    return feature_collection


def istaskrunning(description:str) -> bool:
    task_list = ee.data.getTaskList()
    for t in task_list:
        if t["description"] == description:
            if (t["state"] == "READY") or (t["state"] == "RUNNING"):
                return True
    return False


def mayberun(filename, desc, function, export_task, overwrite=False, dry_run=False, verbose=1,
             bucket_name="worldfloods"):

    if bucket_name is not None:
        fs = fsspec.filesystem("gs", requester_pays=True)
    
        files_in_bucket = fs.glob(f'gs://{bucket_name}/{filename}*')
        if len(files_in_bucket) > 0:
            if overwrite:
                print("\tFile %s exists in the bucket. removing" % filename)
                for b in files_in_bucket:
                    fs.remove(f"gs://{b}")
            else:
                if verbose >= 2:
                    print(f"\tFile {filename} exists in the bucket, it will not be downloaded")
                return
    else:
        files = glob(f"{filename}*")
        if len(files) > 0:
            if overwrite:
                print("\tFile %s exists in the bucket. removing" % filename)
                for b in files:
                    os.remove(b)
            else:
                if verbose >= 2:
                    print(f"\tFile {filename} exists , it will not be downloaded")
                return

    if not dry_run and istaskrunning(desc):
        if verbose >= 2:
            print("\ttask %s already running!" % desc)
        return

    if dry_run:
        print("\tDRY RUN: Downloading file %s" % filename)
        return

    try:
        image_to_download = function()

        if image_to_download is None:
            return

        print("\tDownloading file %s" % filename)

        task = export_task(image_to_download, fileNamePrefix=filename, description=desc)

        task.start()

        return task

    except Exception:
        traceback.print_exc()

    return


def export_task_image(bucket:Optional="worldfloods", crs:str='EPSG:4326',
                      scale:float=10, file_dims=16_384, maxPixels=5_000_000_000) -> Callable:
    """
    function to export images in the WorldFloods format.

    Args:
        bucket:
        scale:
        crs:
        file_dims:
        maxPixels:

    Returns:

    """

    if bucket is not None:
        def export_task(image_to_download, fileNamePrefix, description):
            task = ee.batch.Export.image.toCloudStorage(image_to_download,
                                                        fileNamePrefix=fileNamePrefix,
                                                        description=description,
                                                        crs=crs.upper(),
                                                        skipEmptyTiles=True,
                                                        bucket=bucket,
                                                        scale=scale,
                                                        fileFormat="GeoTIFF",
                                                        formatOptions={"cloudOptimized": True},
                                                        fileDimensions=file_dims,
                                                        maxPixels=maxPixels)
            return task
    else:
        def export_task(image_to_download, fileNamePrefix, description):
            task = ee.batch.Export.image.toDrive(image_to_download,
                                                 fileNamePrefix=fileNamePrefix,
                                                 description=description,
                                                 crs=crs.upper(),
                                                 skipEmptyTiles=True,
                                                 scale=scale,
                                                 fileFormat="GeoTIFF",
                                                 formatOptions={"cloudOptimized": True},
                                                 fileDimensions=file_dims,
                                                 maxPixels=maxPixels)
            return task

    return export_task


def generate_polygon(bbox:Tuple[float, float, float, float]) ->List[List[List[float]]]:
    """
    Generates a list of coordinates: [[x1,y1],[x2,y2],[x3,y3],[x4,y4],[x1,y1]]
    """
    return [[[bbox[0],bbox[1]],
             [bbox[2],bbox[1]],
             [bbox[2],bbox[3]],
             [bbox[0],bbox[3]],
             [bbox[0],bbox[1]]]]


def download_permanent_water(area_of_interest: Polygon, date_search:datetime,
                             path_bucket: str, crs:str='EPSG:4326',
                             name_task:Optional[str]=None, resolution_meters:int=10, requester_pays:bool = True) -> Optional[ee.batch.Task]:
    """
    Downloads yearly permanent water layer from the GEE. (JRC/GSW1_3/YearlyHistory product)

    Args:
        area_of_interest: polygon with the AoI to download
        date_search: start search date
        path_bucket: path in the bucket to export the image. If the files in that bucket exists it does not download
        them.
        crs: crs to export the images. To export them in utm based on location use the `convert_wgs_to_utm` function.
        name_task:
        resolution_meters:

    Returns:
        List of GEE tasks if triggered
    """
    assert path_bucket.startswith("gs://"), f"Path bucket: {path_bucket} must start with gs://"

    fs = fsspec.filesystem("gs", requester_pays = requester_pays)
    filename_full_path = os.path.join(path_bucket, f"{date_search.year}.tif")
    if fs.exists(filename_full_path):
        print(f"File {filename_full_path} exists. It will not be downloaded again")
        return

    ee.Initialize()
    
    path_bucket_no_gs = path_bucket.replace("gs://", "")
    bucket_name = path_bucket_no_gs.split("/")[0]
    path_no_bucket_name = "/".join(path_bucket_no_gs.split("/")[1:])

    area_of_interest_geojson = mapping(area_of_interest)
    pol = ee.Geometry(area_of_interest_geojson)
    bounding_box_aoi = area_of_interest.bounds
    bounding_box_pol = ee.Geometry.Polygon(generate_polygon(bounding_box_aoi))

    if date_search.year >= 2021:
        year = 2021
    else:
        year = date_search.year

    img_export = ee.Image(f"JRC/GSW1_4/YearlyHistory/{year}")

    if name_task is None:
        name_for_desc = os.path.basename(path_no_bucket_name)
    else:
        name_for_desc = name_task

    filename = os.path.join(path_no_bucket_name, f"{year}")
    desc = f"{name_for_desc}_{year}"

    export_task_fun_img = export_task_image(
        bucket=bucket_name,
        crs=crs,
        scale=resolution_meters,
    )

    return mayberun(
        filename,
        desc,
        lambda: img_export.clip(bounding_box_pol),
        export_task_fun_img,
        overwrite=False,
        dry_run=False,
        bucket_name=bucket_name,
        verbose=2,
    )

def download_merit_layer(area_of_interest: Polygon,
                         path_bucket: str, crs:str='EPSG:4326',
                         name_task:Optional[str]=None, resolution_meters:int=10) -> Optional[ee.batch.Task]:
    """
    Downloads MERIT Hydro product ("MERIT/Hydro/v1_0_1") from GEE

    Args:
        area_of_interest: polygon with the AoI to download
        path_bucket: path in the bucket to export the image. If the files in that bucket exists it does not download
        them.
        crs: crs to export the images. To export them in utm based on location use the `convert_wgs_to_utm` function.
        name_task:
        resolution_meters:

    Returns:
        List of GEE tasks if triggered
    """
    assert path_bucket.startswith("gs://"), f"Path bucket: {path_bucket} must start with gs://"

    fs = fsspec.filesystem("gs",requester_pays = True)
    
    filename_full_path = os.path.join(path_bucket, "merit").replace("\\",'/')
    if fs.exists(filename_full_path):
        print(f"File {filename_full_path} exists. It will not be downloaded again")
        return

    ee.Initialize()
    
    path_bucket_no_gs = path_bucket.replace("gs://", "")
    bucket_name = path_bucket_no_gs.split("/")[0]
    path_no_bucket_name = "/".join(path_bucket_no_gs.split("/")[1:])

    area_of_interest_geojson = mapping(area_of_interest)
    bounding_box_aoi = area_of_interest.bounds
    bounding_box_pol = ee.Geometry.Polygon(generate_polygon(bounding_box_aoi))

    img_export = ee.Image("MERIT/Hydro/v1_0_1").float()

    if name_task is None:
        name_for_desc = os.path.basename(path_no_bucket_name)
    else:
        name_for_desc = name_task

    filename = os.path.join(path_no_bucket_name, "merit").replace("\\",'/')
    
    desc = f"{name_for_desc}"

    export_task_fun_img = export_task_image(
        bucket=bucket_name,
        crs=crs,
        scale=resolution_meters,
    )
    

    return mayberun(
        filename,
        desc,
        lambda: img_export.clip(bounding_box_pol),
        export_task_fun_img,
        overwrite=False,
        dry_run=False,
        bucket_name=bucket_name,
        verbose=2,
    )



def convert_wgs_to_utm(lon: float, lat: float) -> str:
    """
    Based on lat and lng, return best utm epsg-code

    https://gis.stackexchange.com/questions/269518/auto-select-suitable-utm-zone-based-on-grid-intersection
    https://stackoverflow.com/questions/40132542/get-a-cartesian-projection-accurate-around-a-lat-lng-pair/40140326#40140326
    Args:
        lon:
        lat:

    Returns: string with the best utm espg-code

    """

    utm_band = str((math.floor((lon + 180) / 6 ) % 60) + 1)
    if len(utm_band) == 1:
        utm_band = '0'+ utm_band
    if lat >= 0:
        epsg_code = 'EPSG:326' + utm_band
        return epsg_code
    epsg_code = 'EPSG:327' + utm_band
    return epsg_code


def process_metadata(path_csv:str, fs=None) -> pd.DataFrame:
    """
    Opens s2info.csv file that are exported in download_s2l89 function. It converts the date fields and
    adds a column indicating which files are available.

    Args:
        path_csv:
        fs:

    Returns:
        dataframe with processed date fields and column indicating if the s2 file is available
    """
    if fs is None:
        fs = fsspec.filesystem("gs", requester_pays=True)
    
    if path_csv.startswith("gs"):
        with fs.open(path_csv, "r") as fh:
            datas2 = pd.read_csv(fh)
    else:
        datas2 = pd.read_csv(path_csv)
        
                         # converters={'datetime': pd.Timestamp})

    datas2["datetime"] = datas2.datetime.apply(lambda x: datetime.fromisoformat(x).replace(tzinfo=timezone.utc))

    datas2["names2file"] = datas2.datetime.apply(lambda x: x.strftime("%Y-%m-%d"))
    datas2["s2available"] = datas2.names2file.apply(lambda x: fs.exists(os.path.join(os.path.dirname(path_csv),
                                                                                     x +".tif")))

    return datas2


def _check_all_downloaded(data:pd.DataFrame,
                          date_start_search: datetime, date_end_search: datetime,
                          filter_s2_fun:Optional[Callable[[pd.DataFrame], pd.Series]],
                          collection_name:str="S2") -> bool:
    """
    True if all images have been downloaded

    Args:
        data:
        date_start_search:
        date_end_search:
        filter_s2_fun:

    Returns:

    """

    min_date = min(data["datetime"])

    if (min_date > date_start_search) and ((min_date-date_start_search).total_seconds() / 3600.) > 10:
        return False

    max_date = max(data["datetime"])
    if (max_date < date_end_search) and ((date_end_search-max_date).total_seconds() / 3600.) > 10:
        return False

    if data.shape[0] <= 0:
        return False

    n_images_col = data.shape[0]
    if filter_s2_fun is not None:
        filter_good = filter_s2_fun(data)
        data = data[filter_good]
        images_available = data["s2available"].sum()
        print(
            f"Found {n_images_col} {collection_name} images between {date_start_search.isoformat()} and {date_end_search.isoformat()} "
            f"{np.sum(filter_good)} satisfies the criteria "
            f"{images_available} are already downloaded")
    else:
        images_available = data["s2available"].sum()
        print(f"Found {n_images_col} {collection_name} images between {date_start_search.isoformat()} and {date_end_search.isoformat()} "
              f"all satisfies the criteria "
              f"{images_available} are already downloaded")

    if data.shape[0] <= 0:
        return False

    if data["s2available"].all():
        return True

    return False


def download_s2l89(area_of_interest: Polygon,
                   date_start_search: datetime, date_end_search: datetime,
                   path_bucket: str,
                   collection_name="COPERNICUS/S2_HARMONIZED", crs:str='EPSG:4326',
                   filter_fun:Callable[[pd.DataFrame], pd.Series]=None,
                   name_task:Optional[str]=None,
                   force_s2cloudless:bool=True,
                   resolution_meters:float=10) -> List[ee.batch.Task]:
    """
    Download time series of S2 or Landsat images between search dates over the given area of interest. It saves the images
    on the path_bucket location. It only downloads images that satisfies the filter_s2 condition.

    Args:
        area_of_interest: shapely polygon with the AoI to download.
        date_start_search: start search date
        date_end_search: end search date
        path_bucket: path in the bucket to export the images. If the files in that bucket exists it does not download
        them.
        collection_name: "COPERNICUS/S2_HARMONIZED" for L1C Sentinel-2 images and ""COPERNICUS/S2_SR_HARMONIZED" for L2A images.
        crs: crs to export the images. To export them in utm based on location use the `convert_wgs_to_utm` function.
        filter_fun: function to filter the images to download. This function receives a dataframe with columns
            "cloud_probability", "valids" and "datetime" the output of this function should be boolean array of the
            with the number of rows of the dataframe that indicates which images of the dataframe to download.
        name_task: if not provided will use the basename of `path_bucket`
        force_s2cloudless:
        resolution_meters: resolution in meters to export the images

    Returns:
        List of running tasks and dataframe with metadata of the S2 files.

    """

    assert path_bucket.startswith("gs://"), f"Path bucket: {path_bucket} must start with gs://"

    path_bucket_no_gs = path_bucket.replace("gs://", "")
    bucket_name = path_bucket_no_gs.split("/")[0]
    path_no_bucket_name = "/".join(path_bucket_no_gs.split("/")[1:])

    fs = fsspec.filesystem("gs", requester_pays = True)

    if collection_name == "Landsat":
        path_csv = os.path.join(path_bucket, "landsatinfo.csv")
    elif collection_name == "S2":
        collection_name = "COPERNICUS/S2_HARMONIZED"
        path_csv = os.path.join(path_bucket, "s2info.csv")
    else:
        path_csv = os.path.join(path_bucket, "s2info.csv")

    if fs.exists(path_csv):
        data = process_metadata(path_csv, fs=fs)
        if _check_all_downloaded(data, date_start_search=date_start_search,
                                 date_end_search=date_end_search,
                                 filter_s2_fun=filter_fun,
                                 collection_name=collection_name):
            return []
        else:
            min_date = min(data["datetime"])
            max_date = max(data["datetime"])
            date_start_search = min(min_date, date_start_search)
            date_end_search = max(max_date, date_end_search)

    ee.Initialize()
    area_of_interest_geojson = mapping(area_of_interest)
    bounding_box_aoi = area_of_interest.bounds
    bounding_box_pol = ee.Geometry.Polygon(generate_polygon(bounding_box_aoi))

    pol = ee.Geometry(area_of_interest_geojson)

    # Grab the S2 images
    if collection_name == "Landsat":
        img_col = get_landsat_collection(date_start_search, date_end_search, pol, verbose=2)
    else:
        img_col = get_s2_collection(date_start_search, date_end_search, pol,
                                    bands=BANDS_NAMES[collection_name],
                                    force_s2cloudless=force_s2cloudless,
                                    collection_name=collection_name, verbose=2)
    if img_col is None:
        return []

    # Get info of the S2 images (convert to table)
    img_col_info_local = image_collection_fetch_metadata(img_col)

    n_images_col = img_col_info_local.shape[0]

    # Save S2 images as csv
    with fs.open(path_csv, "wb") as fh:
        img_col_info_local.to_csv(fh, index=False, mode="wb")

    print(f"Found {n_images_col} {collection_name} images between {date_start_search.isoformat()} and {date_end_search.isoformat()}")

    imgs_list = img_col.toList(n_images_col, 0)

    export_task_fun_img = export_task_image(
        bucket=bucket_name,
        crs=crs,
        scale=resolution_meters,
    )
    if filter_fun is not None:
        filter_good = filter_fun(img_col_info_local)

        if np.sum(filter_good) == 0:
            print("All images are bad")
            return []

        img_col_info_local_good = img_col_info_local[filter_good]
    else:
        img_col_info_local_good = img_col_info_local

    tasks = []
    for good_images in img_col_info_local_good.itertuples():
        img_export = ee.Image(imgs_list.get(good_images.index_image_collection))
        img_export = img_export.select(BANDS_NAMES[collection_name] + ["probability"]).toUint16().clip(bounding_box_pol)

        date = good_images.datetime.strftime('%Y-%m-%d')

        if name_task is None:
            name_for_desc = os.path.basename(path_no_bucket_name)
        else:
            name_for_desc = name_task
        
        filename = os.path.join(path_no_bucket_name, date)
        desc = f"{name_for_desc}_{date}"
        task = mayberun(
            filename,
            desc,
            lambda: img_export,
            export_task_fun_img,
            overwrite=False,
            dry_run=False,
            bucket_name=bucket_name,
            verbose=2,
        )
        if task is not None:
            tasks.append(task)

    return tasks


def image_collection_fetch_metadata(img_col:ee.ImageCollection) -> pd.DataFrame:
    """
    Return the metadata of the provided image collection as a pandas dataframe.
    """
    img_col_info = img_collection_to_feature_collection(img_col,
                                                        ["system:time_start", "valids",
                                                         "cloud_probability"])
    img_col_info_local = gpd.GeoDataFrame.from_features(img_col_info.getInfo())
    img_col_info_local["datetime"] = img_col_info_local["system:time_start"].apply(
        lambda x: datetime.utcfromtimestamp(x / 1000).replace(tzinfo=timezone.utc))
    img_col_info_local["cloud_probability"] /= 100
    img_col_info_local = img_col_info_local[["system:time_start", "valids", "cloud_probability", "datetime"]]
    img_col_info_local["index_image_collection"] = np.arange(img_col_info_local.shape[0])
    return img_col_info_local


def wait_tasks(tasks:List[ee.batch.Task]) -> None:
    """
    Wait for a list of tasks to finish

    Args:
        tasks: list of ee.Tasks

    """
    task_down = []
    for task in tasks:
        if task.active():
            task_down.append((task.status()["description"],task))

    task_error = 0
    while len(task_down) > 0:
        print("%d tasks running" % len(task_down))

        task_down_new = []
        for _i, (t, task) in enumerate(list(task_down)):
            if task.active():
                task_down_new.append((t, task))
                continue
            if task.status()["state"] != "COMPLETED":
                print("Error in task {}:\n {}".format(t, task.status()))
                task_error += 1

        task_down = task_down_new
        time.sleep(60)

    print("Tasks failed: %d" % task_error)


from argparse import ArgumentParser


class UnosatDownloadArgParser(ArgumentParser):

    def __init__(self):
        super().__init__(
            prog="unosat_download.py", description="Download UNOSAT data to buckets"
        )

        self.add_argument(
            "--base-url",
            type=str,
            default="https://unitar.org",
            help="Base URL where UNITAR maps are hosted",
        )

        self.add_argument(
            "--country-list-url",
            type=str,
            default="https://unitar.org/maps/countries",
            help="URL of UNITAR country list for accessing map data",
        )

        self.add_argument(
            "--download-base-regex",
            type=str,
            default="https?://unosat-maps\.web\.cern\.ch/",
            help="RegEx for UNOSAT map download server",
        )

        self.add_argument(
            "--shapefile-bucket",
            type=str,
            default="gs://ml4cc_data_lake/0_DEV/0_Raw/WorldFloods/shapefiles/unosat/",
            help="Path in FS or a major cloud (S3, GCS, Azure Blob) to store shapfiles",
        )

        self.add_argument(
            "--metadata-bucket",
            type=str,
            default="gs://ml4cc_data_lake/0_DEV/0_Raw/WorldFloods/meta/",
            help="Path in FS or a major cloud (S3, GCS, Azure Blob) to store shapfile metadata",
        )




from typing import List

def find_vector(folders_2_unzip: List[str]) -> List[str]:
    """
    This function finds vector folders from a list of folders that had previously
    been unzipped related to Copernicus EMS.
    
    Args:
      folders_2_unzip (List[str]): list of folders
      
    Returns:
      list of folder names that are vector type.
    """
    return [folder for folder in folders_2_unzip if is_vector_file(folder) and not is_rtp_file(folder)]

def is_rtp_file(folder_2_unzip: str) -> bool:
    """
    This function checks if an rtp type file exists in the folder_2_unzip directory.
    
    Args:
      folder_2_unzip (str): name of directory.
      
    Returns:
      boolean
    """
    return 'rtp' in folder_2_unzip.lower()
    
def is_vector_file(folder_2_unzip: str) -> bool:
    """
    This function checks if an vector type file exists in the folder_2_unzip directory.
    
    Args:
      folder_2_unzip (str): name of directory.
      
    Returns:
      boolean
    """
    return 'vector' in folder_2_unzip.lower()


import os
import sys
from pathlib import Path
from typing import List, Optional

from ml4floods.data.utils import download_data_from_bucket, save_file_from_bucket

BUCKET_ID = "ml4floods"
DIR = "worldfloods/public/"


def get_image_path(datclass) -> str:
    """Extracts the S2 Image path

    Args:
        ml_type (str): path to the dataset for the bucket
    Returns:
        path (str): path for the bucket
    """
    return str(Path(datclass.filename))


def download_image(datclass, destination: str, ml_type: str = "train") -> str:
    """Downloads the S2 Image

    Args:
        destination (str): path to where we want to save the data
        ml_type (str): path to the dataset for the bucket
    Returns:
        None
    """
    # get image path
    img_path = get_image_path(datclass)

    # download the image
    return download_data_from_bucket([img_path], destination)


def download_worldfloods_data(
    directories: List[str],
    destination_dir: str,
    ml_split: str = "train",
    bucket_id: Optional[str] = None,
) -> None:
    """Function to download data from the bucket to a local destination directory.
    This function differs from the save_file_from_bucket() function in that
    it takes as input a list of filenames to be downloaded compared to save_file_from_bucket()
    which deals with only a single file.
    Wraps around the save_file_from_bucket() function to download the list of files.

    Args:
        directories (List[str]): List of directories to be downloaded from the bucket.
        destination_dir (str): Path of the destination directory.
        ml_split (str, optional): The split that is to be downloaded.
            Defaults to "train".
            Options: train, val, test
        bucket_id (str, optional): Name of the source GCP bucket.
            Defaults to None.
    """

    if bucket_id is None:
        bucket_id = BUCKET_ID

    for ifile in directories:
        for iprefix in ["S2", "gt"]:

            # where to grab the file
            source = str(Path(DIR).joinpath(ml_split).joinpath(iprefix).joinpath(ifile))
            # Image where to save the file
            destination = Path(destination_dir).joinpath(ml_split).joinpath(iprefix)
            # copy file from bucket to savepath
            save_file_from_bucket(
                bucket_id=BUCKET_ID,
                file_name=source,
                destination_file_path=str(destination),
            )



import argparse
import os

from typing import Optional
import numpy as np
import rasterio
from s2cloudless import S2PixelCloudDetector
from ml4floods.data.worldfloods.configs import BANDS_S2


def sentinel2_to_cloud_mask_preprocess(x):
    """
    takes x in the format of the tif file and rescales it to the format that s2 cloudless expects.
    """
    # last channel is a 'quality assesment' channel rather than a sensor input
    # s2 cloudless also expects channels last and to be scaled to 0-1

    return x[:13, :, :].transpose(1, 2, 0)[None, ...] / 10000


def compute_cloud_mask(x: np.ndarray, threshold: float=0.4, average_over: int=4, dilation_size: int=2, all_bands: bool=True):
    z = sentinel2_to_cloud_mask_preprocess(x)
    cloud_detector = S2PixelCloudDetector(
        threshold=threshold, 
        average_over=average_over, 
        dilation_size=dilation_size, 
        all_bands=all_bands
    )

    cloud_mask = cloud_detector.get_cloud_probability_maps(z)
    cloud_mask = cloud_mask.squeeze()
    return cloud_mask


def compute_cloud_mask_save(cp_path, x, profile):
    cloud_mask = compute_cloud_mask(x)

    if cp_path is not None:
        profile.update(
            count=1,
            compress="lzw",
            dtype="float32",
            driver="COG",
            BIGTIFF="IF_SAFER",
            RESAMPLING="CUBICSPLINE",
        )  # Generate overviews with CUBICSPLINE resampling!

        with rasterio.open(cp_path, "w", **profile) as dst:
            dst.write(cloud_mask.astype(np.float32), 1)

    return cloud_mask


def compute_s2cloudless_probs(s2_image_path: str, window: Optional[rasterio.windows.Window]=None, **kwargs) -> np.ndarray:
    bands_read = list(range(1, len(BANDS_S2) + 1))
    
    # open the S2 Image
    with rasterio.open(s2_image_path, "r") as s2_rst:
        s2_img = s2_rst.read(bands_read, window=window)
        
    return compute_cloud_mask(s2_img, **kwargs)

def create_cloud_mask(tf_path, cp_path, verbose):
    with rasterio.open(tf_path) as x_tif:
        x = x_tif.read()
        compute_cloud_mask_save(cp_path, x, x_tif.profile, verbose)

    return


def main(worldfloods_root, verbose):
    copernicus_s2_path = os.path.join(worldfloods_root, "tiffimages", "S2").replace("\\","/")

    copernicus_cloudprob_path = os.path.join(
        worldfloods_root, "tiffimages", "cloudprob"
    ).replace("\\","/")
    if not os.path.exists(copernicus_cloudprob_path):
        os.mkdir(copernicus_cloudprob_path)

    tiffiles = os.listdir(copernicus_s2_path)

    for i, tf in enumerate(tiffiles):
        tf_path = os.path.join(copernicus_s2_path, tf).replace("\\","/")
        cp_path = os.path.join(copernicus_cloudprob_path, tf).replace("\\","/")
        if i % 20 == 0:
            print("")
        if verbose:
            print("PROCESSING:", tf_path, cp_path)
        if os.path.exists(cp_path):
            #             print(cp_path, "already exists, skipping...")
            print(",", end="", flush=True)
            continue
        try:
            print(".", end="", flush=True)
            create_cloud_mask(tf_path, cp_path, verbose)
        except Exception as e:
            print(f"problem in file {tf_path}: {e}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("worldfloods_root", help="directory where bucket is mounted")
    parser.add_argument(
        "-v", "--verbose", help="turn on verbose mode", action="store_true"
    )

    args = parser.parse_args()

    main(args.worldfloods_root, args.verbose)



import rasterio
import rasterio.windows
import mercantile
from typing import List, Optional, Tuple, Any
from rasterio import warp
import numpy as np

SIZE = 256
WEB_MERCATOR_CRS = "epsg:3857"

PIXEL_PRECISION = 3
# Required because we cant read in parallel with rasterio
def pad_window(window: rasterio.windows.Window, pad_size) -> rasterio.windows.Window:
    """ Add the provided pad to a rasterio window object """
    return rasterio.windows.Window(window.col_off - pad_size[1],
                                   window.row_off - pad_size[0],
                                   width=window.width + 2 * pad_size[1],
                                   height=window.height + 2 * pad_size[1])


def round_outer_window(window:rasterio.windows.Window)-> rasterio.windows.Window:
    """ Rounds a rasterio.windows.Window object to outer (larger) window """
    return window.round_lengths(op="ceil", pixel_precision=PIXEL_PRECISION).round_offsets(op="floor",
                                                                                          pixel_precision=PIXEL_PRECISION)


def read_tile(tiff_path:str, z:int, x:int, y:int, size_output:Tuple[int,int]=(SIZE, SIZE),
              indexes:Optional[List[int]]=None, dst_nodata:Optional[Any]=None,
              resampling=warp.Resampling.cubic_spline) -> Optional[Tuple[np.ndarray, rasterio.transform.Affine]]:
    """
    Reads tile from tileserver location z, x, y from the raster tiff_path. This function reads from the pyramids
    of the tiff file if available.

    Args:
        tiff_path: path to the tiff file
        z: mercantile zoom level
        x: mercantile x coordinate
        y: mercantile y coordinate
        size_output: Size to read
        indexes: 1-based bands to read. input to rasterio read function of the raster.
        dst_nodata:
        resampling: Resampling to use

    Returns:
        numpy array with shape (len(indexes),) + size_output with the data read and affine transformation
    """

    bounds_wgs = mercantile.xy_bounds(int(x), int(y), int(z))

    with rasterio.Env(GDAL_DISABLE_READDIR_ON_OPEN=True, CPL_DEBUG=True), rasterio.open(tiff_path, 'r') as rst:
        x1, y1, x2, y2 = rst.bounds
        rst_bounds = min(x1, x2), min(y1, y2), max(x1, x2), max(y1, y2)
        if rst.crs != WEB_MERCATOR_CRS:
            bounds_read_crs_dest = warp.transform_bounds({"init": WEB_MERCATOR_CRS}, rst.crs,
                                                         left=bounds_wgs.left,
                                                         top=bounds_wgs.top,
                                                         bottom=bounds_wgs.bottom, right=bounds_wgs.right)
        else:
            bounds_read_crs_dest = (bounds_wgs.left, bounds_wgs.bottom, bounds_wgs.right, bounds_wgs.top)

        if rasterio.coords.disjoint_bounds(bounds_read_crs_dest, rst_bounds):
            return None

        rst_transform = rst.transform
        rst_crs = rst.crs
        rst_nodata = rst.nodata or dst_nodata
        rst_dtype = rst.meta["dtype"]

        # Compute the window to read and read the data
        window = rasterio.windows.from_bounds(*bounds_read_crs_dest, rst.transform)
        if rst.crs != WEB_MERCATOR_CRS:
            window = pad_window(window, (1, 1))  # Add padding for bicubic int or for co-registration
            window = round_outer_window(window)

        if indexes is None:
            output_shape_read = (rst.count, size_output[0] + 1, size_output[1] + 1)
            output_shape = (rst.count,) + size_output
        else:
            output_shape_read = (len(indexes), size_output[0] + 1, size_output[1] + 1)
            output_shape = (len(indexes),)+ size_output


        # With out_shape it reads from the pyramids if rst is a COG GeoTIFF
        src_arr = rst.read(indexes=indexes, window=window, out_shape=output_shape_read, boundless=True,
                           fill_value=rst_nodata)

    # Compute transform of readed data and reproject to WEB_MERCATOR_CRS
    input_output_factor = (window.height / output_shape_read[1], window.width / output_shape_read[2])

    transform_window = rasterio.windows.transform(window, rst_transform)
    transform_src = rasterio.Affine(transform_window.a * input_output_factor[1], transform_window.b, transform_window.c,
                                    transform_window.d, transform_window.e * input_output_factor[0], transform_window.f)

    transform_dst = rasterio.transform.from_bounds(west=bounds_wgs.left,north=bounds_wgs.top,
                                                   east=bounds_wgs.right, south=bounds_wgs.bottom,
                                                   width=size_output[0], height=size_output[1])

    rst_arr = np.empty(shape=output_shape, dtype=rst_dtype)

    rst_arr, rst_transform = warp.reproject(source=src_arr, destination=rst_arr,
                                            src_transform=transform_src, src_crs=rst_crs,
                                            src_nodata=rst_nodata,
                                            dst_transform=transform_dst, dst_crs=WEB_MERCATOR_CRS,
                                            dst_nodata=dst_nodata,
                                            resampling=resampling)
    return rst_arr, rst_transform


import requests_html
import re
import time
from dataclasses import dataclass
import json
import geopandas
from fs import open_fs
import requests
import io
from zipfile import ZipFile
import shutil
import numpy as np
import traceback as tb
from ml4floods.data.unosat.unosat_download_arg_parser import UnosatDownloadArgParser

def get_map_links(fetched_page):
    map_links = list(filter(lambda x: re.match(r"^/maps/map/[0-9]+", x), fetched_page.html.links))

    next_page_links = list(filter(lambda x: re.match(r"\?page=[0-9]+", x), fetched_page.html.links))
    next_page_nums = set([int(re.match(r"\?page=([0-9]+)", x).group(1)) for x in next_page_links])
    
    return map_links, next_page_nums
    
def has_shapefile(fetched_page):
    """
    Returns True if a UNITAR product page like https://unitar.org/maps/map/3206 has a shapefile
    """
    return "SHAPEFILE" in fetched_page.content.decode(fetched_page.encoding).upper()
    
def get_flood_shape_link(fetched_page, download_base):
    glide_ids = re.findall(r"GLIDE\:( |&nbsp;)(FL[0-9]+[A-Z]+)", fetched_page.content.decode(fetched_page.encoding))
    if len(glide_ids) > 0:
        # This means we have a GLIDE ID
        assert len(glide_ids) == 1
        shape_links = list(filter(lambda x: re.match(f"{download_base}[a-zA-Z0-9\-/]+{glide_ids[0][1]}_(SHP|shp)\.zip", x), fetched_page.html.links))
        if len(shape_links) > 0:
            assert len(shape_links) == 1
            glide_id = glide_ids[0][1]
            return shape_links[0], glide_id
    
    return None, None
    
def get_country_links_from_url(session, country_list_url):
    fetched_page = session.get(country_list_url)
    return list(filter(lambda x: re.match(r"^/maps/countries/[0-9]+", x), fetched_page.html.links))

def get_map_links_from_country_links_and_url(session, base_url, country_links):
    all_map_links = []
    for country_link in country_links:
        next_page_nums = set([0])
        cur_page = -1
        while cur_page + 1 in next_page_nums:
            cur_page = cur_page + 1

            fetched_page = session.get(base_url + country_link + f"?page={cur_page}")
            map_links, next_page_nums = get_map_links(fetched_page)
            all_map_links.extend(map_links)
            time.sleep(0.2)
    return all_map_links

def get_flood_shape_published_date(fetched_page):
    published_dates = re.findall(r"Published\:( |&nbsp;)([0-9A-Za-z ,]+)", fetched_page.content.decode(fetched_page.encoding))
    if len(published_dates) > 0:
        # This means we have a published date
        assert len(published_dates) == 1
        return published_dates[0][1]
    return None

def get_satellite_source(fetched_page):
    matches = re.findall(r"Satellite Data( |&nbsp;)*(\((Post|Pre|[0-9]+)\))?( |&nbsp;)*\:( |&nbsp;)*([0-9A-Za-z ,\-]+)", fetched_page.content.decode(fetched_page.encoding))
    if len(matches) > 0:
        return matches[0][5]
    return None

def get_satellite_imagery_dates(fetched_page):
    matches = re.findall(r"Imagery Dates( |&nbsp;)*\:( |&nbsp;)*([0-9A-Za-z ,]+)", fetched_page.content.decode(fetched_page.encoding))
    if len(matches) > 0:
        # Sometimes, we see multiple "Imagery Dates" sections (in 50/245 floods as of 02/28/2021)
        # Example: https://unitar.org/maps/map/2562
        return matches[0][2]
    return None

@dataclass
class ShapeFileInfo:
    """Download link and metadata for a shapefile"""
    download_url: str = None
    date_published: str = None
    source_url: str = None
    glide_id: str = None
    satellite_data: str = None
    imagery_dates: str = None
        
    def __hash__(self):
        return self.download_url.__hash__() + self.date_published.__hash__() + self.glide_id.__hash__()

def get_flood_shape_and_meta(session, download_base_regex, base_url, map_link):
    fetched_page = session.get(base_url + map_link)
    if has_shapefile(fetched_page):
        shape_link, glide_id = get_flood_shape_link(fetched_page, download_base_regex)
        if shape_link is not None:
            info = ShapeFileInfo()
            info.date_published = get_flood_shape_published_date(fetched_page)
            info.glide_id = glide_id
            info.download_url = shape_link
            info.source_url = base_url + map_link
            # Get satellite source -- this is formatted in many different ways
            # and needs more work, so will not always be populated. In instances
            # with separate pre-event and post-event imagery, only the first
            # mentioned on the page will be taken.
            info.satellite_data = get_satellite_source(fetched_page)
            # Get imagery dates -- this is not always populated, like satellite data
            # There are many different ways UNOSAT expresses dates over the pages.
            # More work would be helpful here in the long term
            info.imagery_dates = get_satellite_imagery_dates(fetched_page)
            
            return info
    return None

def get_flood_shapefiles(session, base_url, country_list_url, download_base_regex):
    all_country_links = get_country_links_from_url(session, country_list_url)
    all_map_links = get_map_links_from_country_links_and_url(session, base_url, all_country_links)
    flood_shapes = []
    for link in all_map_links:
        fetched_page = session.get(base_url + link)
        info = get_flood_shape_and_meta(session, download_base_regex, base_url, link)
        if info is not None:
            flood_shapes.append(info)
        time.sleep(0.08)
    return flood_shapes


def get_bbox(pd_geo):
    bounds = pd_geo.bounds
    minx = np.min(bounds.minx)
    miny = np.min(bounds.miny)
    maxx = np.max(bounds.maxx)
    maxy = np.max(bounds.maxy)

    return {"west": minx, "east": maxx, "north": maxy, "south": miny}

def produce_metadata_dict(shapefile_info, shapefile_path):
    
    pd_geo = geopandas.read_file(shapefile_path)
    bounding_box = get_bbox(pd_geo)

    if "obj_desc" in pd_geo:
        event_type = np.unique(pd_geo.obj_desc)
        if len(event_type) > 1:
            print("Multiple event types within shapefile {}".format(event_type))
        event_type = event_type[0]
    else:
        event_type = "NaN"
        
    crs_code_space, crs_code = str(pd_geo.crs).split(":")
        
    meta = {
        'event id': shapefile_info.glide_id,
        'layer name': os.path.basename(shapefile_info.download_url),
        'event type': event_type,
        'satellite date': shapefile_info.imagery_dates if shapefile_info.imagery_dates is not None else "NaN",
        'country': "NaN",
        'satellite': shapefile_info.satellite_data if shapefile_info.satellite_data is not None else "NaN",
        'bounding box': bounding_box,
        'reference system': {
            'code space': crs_code_space,
            'code': crs_code
        },
        'abstract': "NaN",
        'purpose': "NaN",
        'source': 'UNOSAT'
    }
    
    return meta

def get_shapefile_from_zip(zip_file):
    
    try:
        shutil.rmtree("./tmp")
    except:
        pass    
    os.mkdir("./tmp")
    
    with ZipFile(zip_file, "r") as f:
        contents = f.namelist()
        # TODO: Some zips only contain WaterExtent but no FloodExtent files -- we should investigate
        # what this means and whether we want to use these too (are they floods?)
        flood_extent_shapefiles = list(filter(lambda x: re.match(r".*\_FloodExtent\_.*\.shp$", x), contents))
        if len(flood_extent_shapefiles) > 0:
            if len(flood_extent_shapefiles) != 1:
                print("WARN: More than one flood shapefile... We will only be using the first one...")
            base_shape_name = os.path.splitext(flood_extent_shapefiles[0])[0]
            relevant_shapefiles = filter(lambda x: x.startswith(base_shape_name), contents)
            # We write all the relevant files out to disk like this because that's the only way I was
            # able to get GeoPandas to accept the file -- it seems not to like a BytesIO object (shapefile)
            # or a list of BytesIO objects (relevant files). It seems to need a path to a .shp file that is
            # located in the same place as its supporting files. If we want to optimize for efficiency,
            # this may be one place to look.
            for relevant_file in relevant_shapefiles:
                try:
                    with f.open(relevant_file, "r") as shp_f, open(f"./tmp/{relevant_file}", "wb") as f_out:
                        f_out.write(shp_f.read())
                except NotImplementedError:
                    print("Unimplemented compression... Skipping...")
                    tb.print_exc()
                    return None
                    

            return "./tmp/" + flood_extent_shapefiles[0]
        return None


def download_shapefiles(shapefile_infos, shapefile_output_dir, metadata_output_dir):
    
    for shapefile_info in shapefile_infos:
        print(shapefile_info)
        shapefiles_and_dbf_zip = requests.get(shapefile_info.download_url).content
        shapefile_path = get_shapefile_from_zip(io.BytesIO(shapefiles_and_dbf_zip))
        if shapefile_path != None:
            meta = produce_metadata_dict(shapefile_info, shapefile_path)
            metadata_name = os.path.splitext(meta["layer name"])[0] + ".json"
            with open_fs(metadata_output_dir +"?strict=False") as fs:
                with fs.open(metadata_name, "w") as f:
                    json.dump(meta, f)

            shapefile_name = shapefile_path.replace("./tmp/", "")
            shapefile_basename = os.path.splitext(shapefile_name)[0]
            layer_basename = os.path.splitext(meta["layer name"])[0]
            for tmpfile in os.listdir("./tmp"):
                target_fname = tmpfile.replace(shapefile_basename, layer_basename)
                with open_fs(os.path.join(shapefile_output_dir) + "?strict=False") as fs:
                    with fs.open(target_fname, "wb") as f, open("./tmp/"+tmpfile, "rb") as f_in:
                        f.write(f_in.read())

def run_unosat_pipeline(base_url, country_list_url, download_base_regex, shapefile_bucket, metadata_bucket):
    session = requests_html.HTMLSession()
    flood_shapefiles = get_flood_shapefiles(session, base_url, country_list_url, download_base_regex)
    download_shapefiles(flood_shapefiles, shapefile_bucket, metadata_bucket)

if __name__ == "__main__":
    args = UnosatDownloadArgParser().parse_args()
    base_url = args.base_url
    country_list_url = args.country_list_url
    download_base_regex = args.download_base_regex
    shapefile_bucket = args.shapefile_bucket
    metadata_bucket = args.metadata_bucket
    run_unosat_pipeline(base_url, country_list_url, download_base_regex, shapefile_bucket, metadata_bucket)




from ml4floods.models.utils.configuration import AttrDict
from ml4floods.data import utils


get_filesystem = utils.get_filesystem
load_json = utils.read_json_from_gcp
save_json = utils.write_json_to_gcp


def setup_config(args) -> AttrDict:
    """
    Loads a config file from disk/bucket. Check channel configuration is consistent and set up the number of channels
    to load the models

    Args:
        args: args to populate the config

    Returns:
        config: config object

    """
    import pprint
    pp = pprint.PrettyPrinter(indent=4)
    
    # 1. Load config json from argparse input
    config = load_json(args.config)
    
    # 2. Add additional fields to config using worldfloods constants etc
    from ml4floods.data.worldfloods.configs import CHANNELS_CONFIGURATIONS

    config['resume_from_checkpoint'] = args.resume_from_checkpoint
    config['gpus'] = args.gpus

    # TODO check channel_configuration is the same in all the parts. Populate this to transforms!
    assert config['model_params']['hyperparameters']['channel_configuration'] ==  config['data_params']['channel_configuration'],\
         f"Set the same channel configuration: {config['model_params']['hyperparameters']['channel_configuration']} {config['data_params']['bands']}"
    
    config['model_params']['hyperparameters']['num_channels'] = len(CHANNELS_CONFIGURATIONS[config['model_params']['hyperparameters']['channel_configuration']])
    
    config = AttrDict.from_nested_dicts(config)

    print('Loaded Config for experiment: ', config.experiment_name)
    pp.pprint(config)
    
    # 3. return config to training
    return config


def get_default_config(config_fp) -> AttrDict:
    """
    Loads a config json file (e.g. from src/models/configurations/worldfloods_template.json) as a config object
    with defaults on gpus

    Args:
        config_fp: path to config file

    Returns:
        config: config object
    """
    import argparse
    parser = argparse.ArgumentParser('WorldFloods')
    parser.add_argument('--config', default=config_fp)
    parser.add_argument('--gpus', default='0', type=str)
    parser.add_argument('--resume_from_checkpoint', default=False, action='store_true')
    # Mode: train, test or deploy
    parser.add_argument('--train', default=False, action='store_true')
    parser.add_argument('--test', default=False, action='store_true')
    parser.add_argument('--deploy', default=False, action='store_true')

    args, _ = parser.parse_known_args()
    
    config = setup_config(args)
    return config



import rasterio
import rasterio.rio.overview
import rasterio.shutil as rasterio_shutil
import os
import tempfile
import numpy as np


from typing import Optional, List

def add_overviews(rst_out, tile_size, verbose=False):
    """ Add overviews to be a cog and be displayed nicely in GIS software """

    overview_level = rasterio.rio.overview.get_maximum_overview_level(*rst_out.shape, tile_size)
    overviews = [2 ** j for j in range(1, overview_level + 1)]

    if verbose:
        print(f"Adding pyramid overviews to raster {overviews}")

    # Copied from https://github.com/cogeotiff/rio-cogeo/blob/master/rio_cogeo/cogeo.py#L274
    rst_out.build_overviews(overviews, rasterio.warp.Resampling.average)
    rst_out.update_tags(ns='rio_overview', resampling='nearest')
    tags = rst_out.tags()
    tags.update(OVR_RESAMPLING_ALG="NEAREST")
    rst_out.update_tags(**tags)
    rst_out._set_all_scales([rst_out.scales[b - 1] for b in rst_out.indexes])
    rst_out._set_all_offsets([rst_out.offsets[b - 1] for b in rst_out.indexes])


def save_cog(out_np: np.ndarray, path_tiff_save: str, profile: dict,
             descriptions:Optional[List[str]] = None,
             tags: Optional[dict] = None,
             dir_tmpfiles:str="."):
    """
    Saves `out_np` np array as a COG GeoTIFF in path_tiff_save. profile is a dict with the geospatial info to be saved
    with the TiFF.

    Args:
        out_np: 3D numpy array to save in CHW format
        path_tiff_save:
        profile: dict with profile to write geospatial info of the dataset: (crs, transform)
        descriptions: List[str]
        tags: extra dict to save as tags
        dir_tmpfiles: dir to create tempfiles if needed

    Returns:
        None

    Examples:
        >> img = np.random.randn(4,256,256)
        >> transform = rasterio.Affine(10, 0, 799980.0, 0, -10, 1900020.0)
        >> save_cog(img, "example.tif", {"crs": {"init": "epsg:32644"}, "transform":transform})
    """

    assert len(out_np.shape) == 3, f"Expected 3d tensor found tensor with shape {out_np.shape}"
    if descriptions is not None:
        assert len(descriptions) == out_np.shape[0], f"Unexpected band descriptions {len(descriptions)} expected {out_np.shape[0]}"

    # Set count, height, width
    for idx, c in enumerate(["count", "height", "width"]):
        if c in profile:
            assert profile[c] == out_np.shape[idx], f"Unexpected shape: {profile[c]} {out_np.shape}"
        else:
            profile[c] = out_np.shape[idx]

    for field in ["crs", "transform"]:
        assert field in profile, f"{field} not in profile: {profile}. it will not write cog without geo information"

    profile["BIGTIFF"] = "IF_SAFER"
    if "dtype" not in profile:
        profile["dtype"] = str(out_np.dtype)
    
    with rasterio.Env() as env:
        cog_driver = "COG" in env.drivers()

    if "RESAMPLING" not in profile:
        profile["RESAMPLING"] = "CUBICSPLINE"  # for pyramids

    if cog_driver:
        # Save tiff locally and copy it to GCP with fsspec is path is a GCP path
        if path_tiff_save.startswith("gs://"):
            import fsspec
            with tempfile.NamedTemporaryFile(dir=dir_tmpfiles, suffix=".tif", delete=True) as fileobj:
                name_save = fileobj.name
        else:
            name_save = path_tiff_save
        profile["driver"] = "COG"
        with rasterio.open(name_save, "w", **profile) as rst_out:
            if tags is not None:
                rst_out.update_tags(**tags)
            rst_out.write(out_np)
            if descriptions is not None:
                for i in range(1, out_np.shape[0] + 1):
                    rst_out.set_band_description(i, descriptions[i-1])

        if path_tiff_save.startswith("gs://"):
            fs = fsspec.filesystem("gs", requester_pays=True)
            fs.put_file(name_save, path_tiff_save)
            # subprocess.run(["gsutil", "-m", "mv", name_save, path_tiff_save])
            if os.path.exists(name_save):
                os.remove(name_save)

        return path_tiff_save

    print("COG driver not available. Generate COG manually with GTiff driver")
    # If COG driver is not available (GDAL < 3.1) we go to copying the file using GTiff driver
    # Set blockysize, blockxsize
    for idx, b in enumerate(["blockysize", "blockxsize"]):
        if b in profile:
            assert profile[b] <= 512, f"{b} is {profile[b]} must be <=512 to be displayed in GEE "
        else:
            profile[b] = min(512, out_np.shape[idx + 1])

    if (out_np.shape[1] >= 512) or (out_np.shape[2] >= 512):
        profile["tiled"] = True

    profile["driver"] = "GTiff"
    with tempfile.NamedTemporaryFile(dir=dir_tmpfiles, suffix=".tif", delete=True) as fileobj:
        named_tempfile = fileobj.name

    with rasterio.open(named_tempfile, "w", **profile) as rst_out:
        if tags is not None:
            rst_out.update_tags(**tags)
        rst_out.write(out_np)
        if descriptions is not None:
            for i in range(1, out_np.shape[0] + 1):
                rst_out.set_band_description(i, descriptions[i - 1])
        
        add_overviews(rst_out, tile_size=profile["blockysize"])
        print("Copying temp file")
        rasterio_shutil.copy(rst_out, path_tiff_save, copy_src_overviews=True, tiled=True,
                             blockxsize=profile["blockxsize"],
                             blockysize=profile["blockysize"],
                             driver="GTiff")

    rasterio_shutil.delete(named_tempfile)
    return path_tiff_save



import datetime
import os
import shutil
import zipfile
from glob import glob
from typing import List, Optional

import geopandas as gpd
import numpy as np
import pandas as pd
from requests_html import HTMLSession
from shapely.ops import unary_union

from ml4floods.data.config import ACCEPTED_FIELDS, RENAME_SATELLITE
import requests
from typing import Dict

from ml4floods.data.utils import filter_land, filter_pols



COPERNICUS_EMS_WEBSCRAPE_DATA  = {"confirmation": 1,
                                  "op":  "Download file",
                                  "form_build_id": "xxxx",
                                  "form_id":"emsmapping_disclaimer_download_form"}


def is_downloadable(url: str) -> bool:
    """
    Function that checks if the url contains a downloadable resource.
    Relies on requests_html package. Helper function for download_vector_cems.

    Args:
      url: A string containing the Copernicus EMS url grabbed
      for specific attributes.

    Returns:
      A boolean indicating if the url is valid
    """

    h = requests.head(url, allow_redirects=True)
    header = h.headers
    content_type = header.get("content-type").lower()
    return "text" not in content_type or "html" not in content_type


def table_floods_ems(event_start_date: str = "2014-05-01") -> pd.DataFrame:
    """
    This function reads the list of EMS Rapid Mapping Activations
    from the Copernicus Emergency Management System for Event type
    'Flood' and 'Storm' and returns a pandas.DataFrame of flood events
    organized in columns {'Code', 'Title', 'CodeDate', 'Type', 'Country'}.

    Args:
      event_start_date (str): The starting date for the span of time until
        the present that the user would like to retrieve flood events for.

    Returns:
      A pandas.DataFrame of Flood events.

    """
    ems_web_page = "https://emergency.copernicus.eu/mapping/list-of-activations-rapid"
    tables = pd.read_html(ems_web_page)[0]
    tables_floods = tables[(tables.Type == "Flood") | (tables.Type == "Storm")]
    tables_floods = tables_floods[tables_floods["Event Date"] >= event_start_date]
    tables_floods = tables_floods.reset_index()[
        ["Act. Code", "Title", "Event Date", "Type", "Country/Terr."]
    ]

    tables_floods = tables_floods.rename(
        {"Act. Code": "Code", "Country/Terr.": "Country", "Event Date": "CodeDate"},
        axis=1,
    )

    return tables_floods.set_index("Code")


def fetch_zip_file_urls(code: str) -> List[str]:
    """
    This Function takes a unique Copernicus EMS hazard activation code and
    retrieves the url which holds the zip files associated with that code.

    e.g. if code = EMSR502, the url holding the zip files of interest is:
    https://emergency.copernicus.eu/mapping/list-of-components/EMSR502/aemfeed

    Args:
      code (str): The unique Copernicus EMS code for an event of interest.

    Returns:
      zip_url_per_code (str): url of the zipfile location for the associated
        input code.
    """
    product_url = "https://emergency.copernicus.eu/mapping/list-of-components/" + code
    session = HTMLSession()
    r = session.get(product_url)

    zip_url_per_code = []
    for zf in r.html.find("a"):
        if (
            ("zip" in zf.attrs["href"])
            and ("REFERENCE_MAP" not in zf.attrs["href"])
            and ("RTP01" not in zf.attrs["href"])
        ):
            zip_url_per_code.append(
                "https://emergency.copernicus.eu" + zf.attrs["href"]
            )

    return zip_url_per_code


def download_vector_cems(zipfile_url:str, folder_out:str="CopernicusEMS") -> str:
    """
    This function downloads the zip files from the zip file url of each
    Copernicus EMS hazard activation and saves them locally to a file
    directory folder_out under a subdirectory based off the activation code.

    Args:
      zipfile_url (str): Url to retrieve the downloadable zip file from
        the Copernicus EMS Rapid Mapping Activations site for a unique hazard
        code of the form 'EMSR[0-9]{3},' where the string EMSR is followed
        by three digits taken from 0-9.

      folder_out (str): The name of the directory created to hold the Copernicus
        EMS zip file download.

    Returns:
      zipfullpath (str): a filepath built from the folder_out handle and the
        Copernicus EMS activation code name embedded in the file.
    """
    # Activation Code derived from zipfile_url
    zipbasename = os.path.basename(zipfile_url)

    # Create a directory called folder_out to hold zipped EMS activations
    if not os.path.exists(folder_out):
        os.mkdir(folder_out)

    # Create a filepath from folder_out and zipbasename if it doesn't exist
    zipfullpath = os.path.join(folder_out, zipbasename)
    if os.path.exists(zipfullpath):
        print("\tFile %s already exists. Not downloaded" % zipfile_url)
        return zipfullpath

    # Check if zipfile_url is valid before making an url request
    if is_downloadable(zipfile_url):
        r = requests.get(zipfile_url, allow_redirects=True)
    else:
        r = requests.post(
            zipfile_url, allow_redirects=True, data=COPERNICUS_EMS_WEBSCRAPE_DATA
        )

    with open(zipfullpath, "wb") as fh:
        fh.write(r.content)

    return zipfullpath


def unzip_copernicus_ems(file_name: str, folder_out: str = "Copernicus_EMS_raw") -> str:
    """
    This function unzips a single zip file from a Copernicus EMS hazard to a
    local directory folder_out with a subdirectory derived from the file handle of
    the file prior to the extension. Copernicus EMS per code zip file may be retrieved
    and downloaded using download_vector_cems locally to access rapid mapping products
    as described in:

    https://emergency.copernicus.eu/mapping/ems/online-manual-rapid-mapping-products

    Args:
      file_name (str): The zip file to be extracted

      folder_out (str): The local directory to store the subdirectories named by the file
        handle naming convention.

    Returns:
       directory_to_extract_to (str): a directory based on the file_name within folder_out.

    """
    # Create directory folder_out locally
    if not os.path.exists(folder_out):
        os.mkdir(folder_out)

    # Create subdirectories to extract zip files to based on file name handle
    zip_ref = zipfile.ZipFile(file_name, "r")
    directory_to_extract_to = os.path.join(
        folder_out, os.path.basename(os.path.splitext(file_name)[0])
    )

    # Check to see if directories exist
    if not os.path.exists(directory_to_extract_to):
        os.mkdir(directory_to_extract_to)
    elif len(os.listdir(directory_to_extract_to)) > 0:
        return directory_to_extract_to

    # Extract
    zip_ref.extractall(directory_to_extract_to)
    zip_ref.close()

    # Remove whitespace with underscores in directory names
    for fextracted in glob(os.path.join(directory_to_extract_to, "*.*")):
        if " " not in fextracted:
            continue
        newname = fextracted.replace(" ", "_")
        shutil.move(fextracted, newname)

    return directory_to_extract_to


# move to config later
formats = [
    "%d/%m/%Y T%H:%M:%SZ",
    "%d/%m/%Y %H:%M:%SZ",
    "%d/%m/%Y T%H:%M:%S",
    "%d/%m/%Y T%H:%MZ",
    "%d/%m/%Y T%H:%M UTC",
    "%d/%m/%Y %H:%M UTC",
    "%d/%m/%Y %H:%M",
    "%d/%m/%Y %H:%M:%S",
    "%Y/%m/%d %H:%M UTC",
    "%d/%m/%Y %H:%M:%S UTC",
    "%m/%d/%Y T%H:%M:%SZ",
]


def is_file_in_directory(parent_dir_of_file: str, file_extension_pattern: str) -> Optional[str]:
    """
    Helper function that checks whether a file already exists in the parent
    directory.

    Args:
      parent_dir_of_file (str): parent directory of the shape files extracted from
        Copernicus EMS.
      file_extension_pattern (str): keyword and file extension for the file of interest.

    Returns:
      A string of the file of interest if it exists in the parent directory, returns
      None otherwise.
    """
    source_file = glob(os.path.join(parent_dir_of_file, file_extension_pattern))
    if len(source_file) == 1:
        return source_file[0]
    elif len(source_file) > 1:
        print(f"Found {len(source_file)} {file_extension_pattern} files in directory. We will not process it {parent_dir_of_file}")
        return
    else:
        print(f"{file_extension_pattern} not found in directory {parent_dir_of_file}")
        return


def post_event_date_difference_is_ok(
    date_post_event: datetime,
    date_ems_code: datetime,
    max_date_post_event: datetime,
    verbose: bool = False,
) -> bool:
    """
    Function to print the datetime difference between pre-flood event and post-flood event.

    Args:
      date_post_event (datetime): date after a flood event.
      date_ems_code (datetime): date Copernicus EMS activation code was issued.
      max_date_post_event (datetime):
      verbose:

    Returns:
      None
    """
    diff_dates = date_post_event - date_ems_code
    if verbose:
        if diff_dates.days < 0:
            print("Difference between dates is negative %d" % diff_dates.days)
            return False

        elif diff_dates.days > 35:
            print("difference too big %d" % diff_dates.days)
            return False

        elif (max_date_post_event - date_post_event).days >= 10:
            print(
                "difference between max date post event and min date post event too big %d"
                % (max_date_post_event - date_post_event).days
            )
            return False

    return True


def _check_hydro_ok(shapefile) -> bool:
    gpd_obj = gpd.read_file(shapefile)

    if gpd_obj.crs is None:
        print(f"{shapefile} file is not georeferenced")
        return False

    if not all(
        notation in ACCEPTED_FIELDS for notation in gpd_obj[COLUMN_W_CLASS_HYDRO]
    ):
        #         print(f"There are unknown fields in the {COLUMN_W_CLASS_HYDRO} column of {shapefile}: {np.unique(gpd_obj[COLUMN_W_CLASS_HYDRO])}")
        not_None_gpd_obj = [
            row for row in gpd_obj[COLUMN_W_CLASS_HYDRO] if row is not None
        ]
        print(
            f"There are unknown fields in the {COLUMN_W_CLASS_HYDRO} column of {shapefile}: {np.unique(np.array(not_None_gpd_obj))}"
        )
        return False

    return True


COLUMN_W_CLASS_OBSERVED_EVENT = "notation"
COLUMN_W_CLASS_HYDRO = "obj_type"


def filter_register_copernicusems(
    unzipped_directory: str, code_date: str, verbose: bool = False
) -> Optional[Dict]:
    """
    Function that collects the following files from the unzipped directories for each Copernicus EMS
    activation code and stores them into a dictionary with additional metadata with respect to the source.
    The files of interest are the shapefiles and associated supporting files for the:

      1. area of interest
      2. hydrography
      3. observed event

    Args:
      unzipped_directory (str): The name of the directory where the extracted EMS files live
      code_date (str): string representation of the EMSR Activation code date of issuance.
      verbose (bool): boolean flag to monitor the difference in pre-event and post-event dates.

    Returns:
      A dictionary with the keys 'name', 'ems_code', 'timestamp', 'satellite', 'area_of_interest'
      'timestamp_ems_code', 'observed_event_file', 'area_of_interest_file.'
      None if there are inconsistencies in the metadata

    """
    # Fetch source files needed to generate floodmap - source, observed event, area of interest
    source_file = is_file_in_directory(unzipped_directory, "*_source*.dbf")
    if not source_file:
        return

    observed_event_file = is_file_in_directory(unzipped_directory, "*_observed*.shp")
    if not observed_event_file:
        return

    area_of_interest_file = is_file_in_directory(unzipped_directory, "*_area*.shp")
    if not area_of_interest_file:
        return

    pd_source = load_source_file(source_file, verbose=verbose)
    if pd_source is None:
        return

    product_name = os.path.basename(observed_event_file).split("_observed")[0]
    ems_code, aoi_code = product_name.split("_")[0:2]

    # Filter content of shapefile
    pd_geo = load_observed_event_file(observed_event_file, verbose=verbose)
    if pd_geo is None:
        return

    # load dmg_src_id fields
    dmg_srd_id_fields = np.unique(pd_geo.dmg_src_id)
    valid_srd_fields_bool = pd_source.src_id.isin(dmg_srd_id_fields)
    if not valid_srd_fields_bool.any():
        if verbose:
            print(f"dmg_srd_id fields not in source file {dmg_srd_id_fields}")
        return 

    min_date_post_event = min(pd_source.loc[valid_srd_fields_bool & (pd_source.eventphase == "Post-event"), "date"])
    max_date_post_event = max(pd_source.loc[valid_srd_fields_bool & (pd_source.eventphase == "Post-event"), "date"])

    satellite_post_event = pd_source.loc[(pd_source.eventphase == "Post-event") & (pd_source.date == max_date_post_event), "source_nam"].iloc[0]

    date_ems_code = datetime.datetime.strptime(code_date, "%Y-%m-%d").replace(
        tzinfo=datetime.timezone.utc
    )

    if not post_event_date_difference_is_ok(
        min_date_post_event, date_ems_code, max_date_post_event, verbose
    ):
        return

    # Check if pre-event date precedes post-event date
    content_pre_event = {}
    if np.any(pd_source.eventphase == "Pre-event"):
        date_pre_event = max(
            np.array(pd_source[pd_source.eventphase == "Pre-event"]["date"])
        )
        satellite_pre_event = np.array(
            pd_source[
                (pd_source.eventphase == "Pre-event")
                & (pd_source.date == date_pre_event)
            ]["source_nam"]
        )[0]

        content_pre_event["satellite_pre_event"] = satellite_pre_event
        content_pre_event["timestamp_pre_event"] = date_pre_event
        if (min_date_post_event - date_pre_event).days < 0 and verbose:
            print(
                "Date pre event %s is after date post event %s"
                % (
                    date_pre_event.strftime("%Y-%m-%d"),
                    min_date_post_event.strftime("%Y-%m-%d"),
                )
            )
            return

    if not isinstance(satellite_post_event, str) and verbose:
        print("Satellite post event: {} is not a string!".format(satellite_post_event))
        return

    if satellite_post_event in RENAME_SATELLITE:
        satellite_post_event = RENAME_SATELLITE[satellite_post_event]

    area_of_interest = gpd.read_file(area_of_interest_file)

    if area_of_interest.crs is None:
        print(f"{area_of_interest_file} file is not georeferenced")
        return

    area_of_interest_crs = str(area_of_interest.crs)

    if area_of_interest_crs.lower() != "epsg:4326":
        area_of_interest.to_crs(crs="epsg:4326", inplace=True)

    # Save pol of area of interest in epsg:4326 (lat/lng)
    area_of_interest_pol = unary_union(area_of_interest["geometry"])

    if "obj_desc" in pd_geo:
        event_type = np.unique(pd_geo.obj_desc)
        if len(event_type) > 1:
            print("Multiple event types within shapefile {}".format(event_type))
        event_type = event_type[0]
    else:
        event_type = "NaN"

    register = {
        "event id": product_name,
        "layer name": os.path.basename(os.path.splitext(observed_event_file)[0]),
        "event type": event_type,
        "satellite date": max_date_post_event,
        "country": "NaN",
        "satellite": satellite_post_event,
        "bounding box": get_bbox(pd_geo),
        "reference system": area_of_interest_crs,
        "abstract": "NaN",
        "purpose": "NaN",
        "source": "CopernicusEMS",
        "area_of_interest_polygon": area_of_interest_pol,
        # CopernicusEMS specific fields
        "observed_event_file": os.path.basename(observed_event_file),
        "area_of_interest_file": os.path.basename(area_of_interest_file),
        "ems_code": ems_code,
        "aoi_code": aoi_code,
        "date_ems_code": date_ems_code
    }

    register.update(content_pre_event)

    # Add hydrography polygons and hydrography lines
    name_possibilities = ["_hydrographyA_", "_hydrography_a", "_hydrography_p"]
    for name_pos in name_possibilities:
        hydrology_file_as = glob(os.path.join(unzipped_directory, f"*{name_pos}*.shp"))
        if len(hydrology_file_as) == 1:
            if not _check_hydro_ok(hydrology_file_as[0]):
                return
            register["hydrology_file"] = os.path.basename(hydrology_file_as[0])

    name_possibilities = ["_hydrographyL_", "_hydrography_l"]
    for name_pos in name_possibilities:
        hydrology_file_l = glob(os.path.join(unzipped_directory, f"*{name_pos}*.shp"))
        if len(hydrology_file_l) == 1:
            if not _check_hydro_ok(hydrology_file_l[0]):
                return
            register["hydrology_file_l"] = os.path.basename(hydrology_file_l[0])

    return register

def load_observed_event_file(observed_event_file:str, verbose:bool=False) -> Optional[gpd.GeoDataFrame]:
    pd_geo = gpd.read_file(observed_event_file)
    if np.any(pd_geo["event_type"] != "5-Flood") and verbose:
        print(
            f"{observed_event_file} Event type is not Flood {np.unique(pd_geo['event_type'][~pd_geo.event_type.isna()])}"
        )
        return

    if pd_geo.notation.isna().any():
        if verbose:
            print(f"Found na in field {COLUMN_W_CLASS_OBSERVED_EVENT}. Replacing them with 'Flooded area'")
        pd_geo.loc[pd_geo[COLUMN_W_CLASS_OBSERVED_EVENT].isna(), COLUMN_W_CLASS_OBSERVED_EVENT] = 'Flooded area'

    if not all(
        notation in ACCEPTED_FIELDS
        for notation in pd_geo[COLUMN_W_CLASS_OBSERVED_EVENT]
    ):
        print(
            f"There are unknown fields in the {COLUMN_W_CLASS_OBSERVED_EVENT} column of {observed_event_file}: {np.unique(pd_geo[COLUMN_W_CLASS_OBSERVED_EVENT])}"
        )
        return

    if pd_geo.crs is None:
        print(f"{observed_event_file} file is not georeferenced")
        return
    return pd_geo


def parse_date_messy(date_list):
    """
    Helper function for load_source_file
    """
    for fm in formats:
        try:
            # date_list[0] = date_list[0].replace("119/09", "19/09").replace("05/24", "24/05")
            time_part = (
                date_list[1].replace(".", ":").replace(";", ":").replace("UCT", "UTC")
            )
            if time_part == "Not Applicable":
                time_part = "00:00"
            date_post_event = datetime.datetime.strptime(
                date_list[0] + " " + time_part, fm
            )
            if date_post_event.tzname() is None:
                date_post_event = date_post_event.replace(tzinfo=datetime.timezone.utc)
            return date_post_event
        except ValueError:
            pass

    return None


def load_source_file(source_file, filter_event_dates=True, verbose=False):
    """
    Helper function for filter_register_copernicusems
    """

    pd_source = gpd.read_file(source_file)

    # pd_source = Dbf5(source_file).to_dataframe()

    if "eventphase" not in pd_source.columns or not np.any(
        pd_source.eventphase == "Post-event"
    ):
        if verbose:
            print("\t eventphase not found or not Post-event in source file")
        return

    dates_formated = []
    for d in pd_source.itertuples():
        date = parse_date_messy([d.src_date, d.source_tm])
        dates_formated.append(date)
    pd_source["date"] = dates_formated
    if filter_event_dates:
        pd_source = pd_source[
            (pd_source.eventphase == "Post-event")
            | (pd_source.eventphase == "Pre-event")
        ]

        if np.any(pd.isna(pd_source.date)):
            if verbose:
                print("\t There are event phase files with undefined dates")

            pd_source = pd_source[~pd.isna(pd_source.date)]

        if not np.any(pd_source.eventphase == "Post-event"):
            if verbose:
                print("\t Post-event removed after filtering")
            return

    return pd_source


def generate_floodmap(
    metadata_floodmap: Dict, folder_files: str, filterland: bool = True
) -> gpd.GeoDataFrame:
    """ Generates a floodmap with the joined info of the hydro and flood. """

    area_of_interest = gpd.read_file(
        os.path.join(
            folder_files,
            metadata_floodmap["area_of_interest_file"],
        )
    )

    crs = area_of_interest.crs

    area_of_interest_pol = unary_union(area_of_interest["geometry"])

    mapdf = filter_pols(
        gpd.read_file(
            os.path.join(
                folder_files,
                metadata_floodmap["observed_event_file"],
            )
        ).to_crs(crs),
        area_of_interest_pol,
    )
    assert mapdf.shape[0] > 0, f"No polygons within bounds for {metadata_floodmap}"

    floodmap = gpd.GeoDataFrame({"geometry": mapdf.geometry}, crs=mapdf.crs)
    floodmap["w_class"] = mapdf[COLUMN_W_CLASS_OBSERVED_EVENT]
    floodmap["source"] = "flood"

    if "hydrology_file" in metadata_floodmap:
        mapdf_hydro = filter_pols(
            gpd.read_file(
                os.path.join(
                    folder_files,
                    metadata_floodmap["hydrology_file"],
                )
            ).to_crs(crs),
            area_of_interest_pol,
        )
        mapdf_hydro = (
            filter_land(mapdf_hydro)
            if filterland and (mapdf_hydro.shape[0] > 0)
            else mapdf_hydro
        )
        if mapdf_hydro.shape[0] > 0:
            mapdf_hydro["source"] = "hydro"
            mapdf_hydro = mapdf_hydro.rename({COLUMN_W_CLASS_HYDRO: "w_class"}, axis=1)
            mapdf_hydro = mapdf_hydro[["geometry", "w_class", "source"]].copy()
            floodmap = pd.concat([floodmap, mapdf_hydro], axis=0, ignore_index=True)

    # Add "hydrology_file_l"?? must be handled in later in create_gt.compute_water function
    if "hydrology_file_l" in metadata_floodmap:
        mapdf_hydro = filter_pols(
            gpd.read_file(
                os.path.join(
                    folder_files,
                    metadata_floodmap["hydrology_file_l"],
                )
            ).to_crs(crs),
            area_of_interest_pol,
        )
        if mapdf_hydro.shape[0] > 0:
            mapdf_hydro["source"] = "hydro_l"
            mapdf_hydro = mapdf_hydro.rename({COLUMN_W_CLASS_HYDRO: "w_class"}, axis=1)
            mapdf_hydro = mapdf_hydro[["geometry", "w_class", "source"]].copy()
            floodmap = pd.concat([floodmap, mapdf_hydro], axis=0, ignore_index=True)

    # Concat area of interest
    area_of_interest["source"] = "area_of_interest"
    area_of_interest["w_class"] = "area_of_interest"
    area_of_interest = area_of_interest[["geometry", "w_class", "source"]].copy()
    floodmap = pd.concat([floodmap, area_of_interest], axis=0, ignore_index=True)

    assert floodmap.crs is not None, "Unexpected error. floodmap is not georreferenced!"

    floodmap.loc[floodmap.w_class.isna(), "w_class"] = "Not Applicable"

    return floodmap


def generate_polygon(bbox):
    """
    Generates a list of coordinates: [[x1,y1],[x2,y2],[x3,y3],[x4,y4],[x1,y1]]
    """
    return [
        [bbox[0], bbox[1]],
        [bbox[2], bbox[1]],
        [bbox[2], bbox[3]],
        [bbox[0], bbox[3]],
        [bbox[0], bbox[1]],
    ]


def get_bbox(pd_geo: gpd.GeoDataFrame) -> Dict:
    """
    This function is a helper function of processing_register
    This function takes a geopandas dataframe

    """
    bounds = pd_geo.bounds
    minx = np.min(bounds.minx)
    miny = np.min(bounds.miny)
    maxx = np.max(bounds.maxx)
    maxy = np.max(bounds.maxy)

    return {"west": minx, "east": maxx, "north": maxy, "south": miny}


COPY_FORMATS = ["dbf", "prj", "shp", "shx"]



import numbers

from ml4floods.preprocess.tiling import WindowSlices
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import rasterio
import rasterio.windows
import torch
from torch.utils.data import Dataset
import contextlib
from ml4floods.data import utils


from ml4floods.data.worldfloods.configs import BANDS_S2

import threading


class WorldFloodsDataset(Dataset):
    """A prepackaged WorldFloods PyTorch Dataset
    This initializes the dataset given a set a set of image files with a
    subdirectory for the training and testing data ("image_prefix" and "gt_prefix").

    Args:
        image_files (List[str]): the image files to be loaded into the
            dataset
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        transforms (Callable): the transformations used within the
            training data module

    Attributes:
        image_files (List[str]): the image files to be loaded into the
            dataset
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        transforms (Callable): the transformations used within the
            training data module
        bands: List[int]
            0-based list of bands to read from BANDS_S2
    """

    def __init__(
        self,
        image_files: List[str],
        image_prefix: str = "/image_files/",
        gt_prefix: str = "/gt_files/",
        transforms: Optional[Callable] = None,
        bands: List[int] = list(range(len(BANDS_S2))),
        lock_read: bool = False,
    ) -> None:

        self.image_files = image_files
        self.image_prefix = image_prefix
        self.gt_prefix = gt_prefix
        self.transforms = transforms
        self.bands_read = bands
        if lock_read:
            self._lock = threading.Lock()
        else:
            self._lock = contextlib.nullcontext()

        # sort to make sure that the order is deterministic
        # (order of the flow of data points to the ML model)
        # TODO: Do this for the list of filepaths at the end as well
        self.image_files.sort()

    def __getitem__(self, idx: int) -> Dict:
        """Index to select an image

        Args:
            idx (int): index

        Returns:
            a dictionary with the image and mask keys
            {"image", "mask"}
        """

        # get filenames
        image_name = self.image_files[idx]

        y_name = image_name.replace(self.image_prefix, self.gt_prefix, 1)

        image_tif = rasterio_read(
            image_name, self._lock, channels=[c + 1 for c in self.bands_read]
        )

        mask_tif = rasterio_read(y_name, self._lock)

        # get rid of nan, convert to float
        image = np.nan_to_num(image_tif).astype(np.float32)

        # The 0-index comes from reading all the bands with f.read()
        mask = np.nan_to_num(mask_tif)

        # Apply transformation
        if self.transforms is not None:
            data = self.transforms(image=image, mask=mask)
        else:
            data = {"image": image, "mask": mask}
        # return x, y
        return data

    def __len__(self) -> int:
        return len(self.image_files)


class WorldFloodsDatasetTiled(Dataset):
    """A prepackaged WorldFloods PyTorch Dataset
    This initializes the dataset given a set a set of image files with a
    subdirectory for the training and testing data ("image_prefix" and "gt_prefix").
    This also does the tiling under the hood given the windowsize.

    Args:
        list_of_windows (List[WindowSlices]):  a list of
            namedtuples each consisting of a filename and a rasterio.window
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        transforms (Callable): the transformations used within the
            training data module

    Attributes:
        list_of_windows (List[WindowSlices]):  a list of
            namedtuples each consisting of a filename and a rasterio.window
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        transforms (Callable): the transformations used within the
            training data module
        bands: List[int]
            0-based list of bands to read from BANDS_S2
    """

    def __init__(
        self,
        list_of_windows: List[WindowSlices],
        image_prefix: str = "/image_files/",
        gt_prefix: str = "/gt_files/",
        transforms: Optional[Callable] = None,
        bands: List[int] = list(range(len(BANDS_S2))),
        lock_read: bool = False,
    ) -> None:

        self.image_prefix = image_prefix
        self.gt_prefix = gt_prefix
        self.transforms = transforms
        self.channels_read = bands

        if lock_read:
            # Useful when reading from bucket
            self._lock = threading.Lock()
        else:
            self._lock = contextlib.nullcontext()

        self.list_of_windows = list_of_windows

    def get_label(self, idx: int) -> np.ndarray:
        """
        Method to read only the label. This function is useful for filtering the patches of the Dataset
        Args:
            idx:

        Returns:


        """
        sub_window = self.list_of_windows[idx]
        y_name = sub_window.file_name.replace(self.image_prefix, self.gt_prefix, 1)
        return rasterio_read(
            y_name,
            self._lock,
            channels=None,
            kwargs_rasterio={
                "window": sub_window.window,
                "boundless": True,
                "fill_value": 0,
            },
        )

    def __getitem__(self, idx: int) -> Dict:
        """Index to select an image tile

        Args:
            idx (int): index

        Returns:
            a dictionary with the keys to the image tiles and mask tiles
            {"image", "mask"}
        """
        # get filenames from named tuple
        sub_window = self.list_of_windows[idx]

        # get filename
        image_name = sub_window.file_name

        # replace string for image_prefix
        image_name = image_name.replace(self.gt_prefix, self.image_prefix, 1)

        # replace string for gt_prefix
        y_name = image_name.replace(self.image_prefix, self.gt_prefix, 1)

        # Open Image File
        image_tif = rasterio_read(
            image_name,
            self._lock,
            channels=[c + 1 for c in self.channels_read],
            kwargs_rasterio={
                "window": sub_window.window,
                "boundless": True,
                "fill_value": 0,
            },
        )
        mask_tif = rasterio_read(
            y_name,
            self._lock,
            channels=None,
            kwargs_rasterio={
                "window": sub_window.window,
                "boundless": True,
                "fill_value": 0,
            },
        )

        # get rid of nan, convert to float
        image = np.nan_to_num(image_tif).astype(np.float32)

        mask = np.nan_to_num(mask_tif).astype(int)

        # Apply transformation
        if self.transforms is not None:
            data = self.transforms(image=image, mask=mask)
        else:
            data = {"image": image, "mask": mask}
        # return x, y
        return data

    def __len__(self) -> int:
        return len(self.list_of_windows)


def rasterio_read(
    image_name: str, lock, channels: List[int] = None, kwargs_rasterio: Dict = {}
) -> np.ndarray:
    with lock:
        with utils.rasterio_open_read(image_name) as f:
            im_tif = f.read(channels, **kwargs_rasterio)

    return im_tif


def load_input(tiff_input:str, channels:Union[List[int],List[str]],
               window:Optional[rasterio.windows.Window]=None) -> Tuple[torch.Tensor, rasterio.transform.Affine]:
    """
    Reads from a tiff the specified channel and window.

    Args:
        tiff_input: path to geotiff file
        window: rasterio.Window object to read (None to read all)
        channels: 0-based channels to read or names of the bands (we will use `.descriptions` to find the band
            names of the `tiff_input`).

    Returns:
        3-D tensor (len(channels), H, W), Affine transform to geo-reference the array read.

    """
    with utils.rasterio_open_read(tiff_input) as rst:
        if isinstance(channels[0], numbers.Number):
            indexes = (np.array(channels) + 1).tolist()
        else:
            channels_tiff = list(rst.descriptions)
            indexes = [channels_tiff.index(c) + 1 for c in channels]

        inputs = rst.read(indexes, window = window)

        # Shifted transform based on the given window (used for plotting)
        transform = rst.transform if window is None else rasterio.windows.transform(window, rst.transform)
        torch_inputs = torch.tensor(np.nan_to_num(inputs).astype(np.float32))
    return torch_inputs, transform